import logging
import os
import platform
import re
import sys
from collections import defaultdict
from datetime import datetime
from difflib import SequenceMatcher
from threading import Lock
from typing import TYPE_CHECKING, Any, Dict, Generator, List, Optional, Tuple, Union

import click
from click import BadParameter
from dparse import filetypes, parse
from packaging.specifiers import SpecifierSet
from packaging.utils import canonicalize_name
from packaging.version import parse as parse_version
from requests import PreparedRequest
from ruamel.yaml import YAML
from ruamel.yaml.error import MarkedYAMLError
from safety_schemas.models import TelemetryModel

from safety.constants import (
    HASH_REGEX_GROUPS,
    SYSTEM_CONFIG_DIR,
    USER_CONFIG_DIR,
)
from safety.errors import InvalidProvidedReportError
from safety.events.event_bus import start_event_bus
from safety.models import (
    Package,
    RequirementFile,
    SafetyRequirement,
    is_pinned_requirement,
)


if TYPE_CHECKING:
    from safety.cli_util import CustomContext
    from safety.models import SafetyCLI
    from safety.auth.models import Auth
    from safety.auth.utils import SafetyAuthSession

    import typer


LOG = logging.getLogger(__name__)


def is_a_remote_mirror(mirror: str) -> bool:
    """
    Check if a mirror URL is remote.

    Args:
        mirror (str): The mirror URL.

    Returns:
        bool: True if the mirror URL is remote, False otherwise.
    """
    return mirror.startswith("http://") or mirror.startswith("https://")


def is_supported_by_parser(path: str) -> bool:
    """
    Check if the file path is supported by the parser.

    Args:
        path (str): The file path.

    Returns:
        bool: True if the file path is supported, False otherwise.
    """
    supported_types = (
        ".txt",
        ".in",
        ".yml",
        ".ini",
        "Pipfile",
        "Pipfile.lock",
        "setup.cfg",
        "poetry.lock",
    )
    return path.endswith(supported_types)


def parse_requirement(dep: Any, found: Optional[str]) -> SafetyRequirement:
    """
    Parse a requirement.

    Args:
        dep (Any): The dependency.
        found (str): The location where the dependency was found.

    Returns:
        SafetyRequirement: The parsed requirement.
    """
    req = SafetyRequirement(dep)
    req.found = found

    if req.specifier == SpecifierSet(""):
        req.specifier = SpecifierSet(">=0")

    return req


def find_version(requirements: List[SafetyRequirement]) -> Optional[str]:
    """
    Find the version of a requirement.

    Args:
        requirements (List[SafetyRequirement]): The list of requirements.

    Returns:
        Optional[str]: The version if found, None otherwise.
    """
    ver = None

    if len(requirements) != 1:
        return ver

    specs = requirements[0].specifier

    if is_pinned_requirement(specs):
        ver = next(iter(requirements[0].specifier)).version

    return ver


def read_requirements(fh: Any, resolve: bool = True) -> Generator[Package, None, None]:  # type: ignore
    """
    Reads requirements from a file-like object and (optionally) from referenced files.

    Args:
        fh (Any): The file-like object to read from.
        resolve (bool): Resolves referenced files.

    Returns:
        Generator: Yields Package objects.
    """
    is_temp_file = not hasattr(fh, "name")
    path = None
    found = "temp_file"
    file_type = filetypes.requirements_txt
    absolute_path: Optional[str] = None

    if not is_temp_file and is_supported_by_parser(fh.name):
        LOG.debug("not temp and a compatible file")
        path = fh.name
        absolute_path = os.path.abspath(path)
        SafetyContext().scanned_full_path.append(absolute_path)
        found = path
        file_type = None

    LOG.debug(f"Path: {path}")
    LOG.debug(f"File Type: {file_type}")
    LOG.debug("Trying to parse file using dparse...")
    content = fh.read()
    LOG.debug(f"Content: {content}")
    dependency_file = parse(content, path=path, resolve=resolve, file_type=file_type)
    LOG.debug(f"Dependency file: {dependency_file.serialize()}")
    LOG.debug(
        f"Parsed, dependencies: {[dep.serialize() for dep in dependency_file.resolved_dependencies]}"
    )

    reqs_pkg = defaultdict(list)

    for req in dependency_file.resolved_dependencies:
        reqs_pkg[canonicalize_name(req.name)].append(req)

    for pkg, reqs in reqs_pkg.items():
        requirements = list(
            map(lambda req: parse_requirement(req, absolute_path), reqs)
        )
        version = find_version(requirements)

        yield Package(
            name=pkg,
            version=version,
            requirements=requirements,
            found=found,
            absolute_path=absolute_path,
            insecure_versions=[],
            secure_versions=[],
            latest_version=None,
            latest_version_without_known_vulnerabilities=None,
            more_info_url=None,
        )


def get_proxy_dict(
    proxy_protocol: str, proxy_host: str, proxy_port: int
) -> Optional[Dict[str, str]]:
    """
    Get the proxy dictionary for requests.

    Args:
        proxy_protocol (str): The proxy protocol.
        proxy_host (str): The proxy host.
        proxy_port (int): The proxy port.

    Returns:
        Optional[Dict[str, str]]: The proxy dictionary if all parameters are provided, None otherwise.
    """
    if proxy_protocol and proxy_host and proxy_port:
        # Safety only uses https request, so only https dict will be passed to requests
        return {"https": f"{proxy_protocol}://{proxy_host}:{str(proxy_port)}"}
    return None


def get_license_name_by_id(license_id: int, db: Dict[str, Any]) -> Optional[str]:
    """
    Get the license name by its ID.

    Args:
        license_id (int): The license ID.
        db (Dict[str, Any]): The database containing license information.

    Returns:
        Optional[str]: The license name if found, None otherwise.
    """
    licenses = db.get("licenses", [])
    for name, id in licenses.items():
        if id == license_id:
            return name
    return None


def get_flags_from_context() -> Dict[str, str]:
    """
    Get the flags from the current click context.

    Returns:
        Dict[str, str]: A dictionary of flags and their corresponding option names.
    """
    flags = {}
    context = click.get_current_context(silent=True)

    if context:
        for option in context.command.params:
            flags_per_opt = option.opts + option.secondary_opts
            for flag in flags_per_opt:
                flags[flag] = option.name

    return flags


def get_used_options() -> Dict[str, Dict[str, int]]:
    """
    Get the used options from the command-line arguments.

    Returns:
        Dict[str, Dict[str, int]]: A dictionary of used options and their counts.
    """
    flags = get_flags_from_context()
    used_options = {}

    for arg in sys.argv:
        cleaned_arg = arg if "=" not in arg else arg.split("=")[0]
        if cleaned_arg in flags:
            option_used = flags.get(cleaned_arg)

            if option_used in used_options:
                used_options[option_used][cleaned_arg] = (
                    used_options[option_used].get(cleaned_arg, 0) + 1
                )
            else:
                used_options[option_used] = {cleaned_arg: 1}

    return used_options


def get_version() -> str:
    """
    Get the version of Safety.

    Returns:
        str: The Safety version.
    """
    from importlib.metadata import version

    return version("safety")


def get_primary_announcement(
    announcements: List[Dict[str, Any]],
) -> Optional[Dict[str, Any]]:
    """
    Get the primary announcement from a list of announcements.

    Args:
        announcements (List[Dict[str, Any]]): The list of announcements.

    Returns:
        Optional[Dict[str, Any]]: The primary announcement if found, None otherwise.
    """
    for announcement in announcements:
        if announcement.get("type", "").lower() == "primary_announcement":
            try:
                from safety.output_utils import build_primary_announcement

                build_primary_announcement(announcement, columns=80)
            except Exception as e:
                LOG.debug(f"Failed to build primary announcement: {str(e)}")
                return None

            return announcement

    return None


def get_basic_announcements(
    announcements: List[Dict[str, Any]], include_local: bool = True
) -> List[Dict[str, Any]]:
    """
    Get the basic announcements from a list of announcements.

    Args:
        announcements (List[Dict[str, Any]]): The list of announcements.
        include_local (bool): Whether to include local announcements.

    Returns:
        List[Dict[str, Any]]: The list of basic announcements.
    """
    return [
        announcement
        for announcement in announcements
        if announcement.get("type", "").lower() != "primary_announcement"
        and not announcement.get("local", False)
        or (announcement.get("local", False) and include_local)
    ]


def filter_announcements(
    announcements: List[Dict[str, Any]], by_type: str = "error"
) -> List[Dict[str, Any]]:
    """
    Filter announcements by type.

    Args:
        announcements (List[Dict[str, Any]]): The list of announcements.
        by_type (str): The type of announcements to filter by.

    Returns:
        List[Dict[str, Any]]: The filtered announcements.
    """
    return [
        announcement
        for announcement in announcements
        if announcement.get("type", "").lower() == by_type
    ]


def build_telemetry_data(
    telemetry: bool = True,
    command: Optional[str] = None,
    subcommand: Optional[str] = None,
) -> TelemetryModel:
    """Build telemetry data for the Safety context.

    Args:
        telemetry (bool): Whether telemetry is enabled.
        command (Optional[str]): The command.
        subcommand (Optional[str]): The subcommand.

    Returns:
        TelemetryModel: The telemetry data model.
    """
    context = SafetyContext()

    body = (
        {
            "os_type": os.environ.get("SAFETY_OS_TYPE", None) or platform.system(),
            "os_release": os.environ.get("SAFETY_OS_RELEASE", None)
            or platform.release(),
            "os_description": os.environ.get("SAFETY_OS_DESCRIPTION", None)
            or platform.platform(),
            "python_version": platform.python_version(),
            "safety_command": command if command else context.command,
            "safety_options": get_used_options(),
        }
        if telemetry
        else {}
    )

    body["safety_version"] = get_version()
    body["safety_source"] = (
        os.environ.get("SAFETY_SOURCE", None) or context.safety_source
    )

    if "safety_options" not in body:
        body["safety_options"] = {}

    LOG.debug(f"Telemetry body built: {body}")

    return TelemetryModel(**body)


def build_git_data() -> Dict[str, Any]:
    """Build git data for the repository.

    Returns:
        Dict[str, str]: The git data.
    """
    import subprocess

    def git_command(commandline: List[str]) -> str:
        return (
            subprocess.run(
                commandline, stdout=subprocess.PIPE, stderr=subprocess.DEVNULL
            )
            .stdout.decode("utf-8")
            .strip()
        )

    try:
        is_git = git_command(["git", "rev-parse", "--is-inside-work-tree"])
    except Exception:
        is_git = False

    if is_git == "true":
        result = {"branch": "", "tag": "", "commit": "", "dirty": "", "origin": ""}

        try:
            result["branch"] = git_command(
                ["git", "symbolic-ref", "--short", "-q", "HEAD"]
            )
            result["tag"] = git_command(["git", "describe", "--tags", "--exact-match"])

            commit = git_command(
                ["git", "describe", '--match=""', "--always", "--abbrev=40", "--dirty"]
            )
            result["dirty"] = str(commit.endswith("-dirty"))
            result["commit"] = commit.split("-dirty")[0]

            result["origin"] = git_command(["git", "remote", "get-url", "origin"])
        except Exception:
            pass

        return result
    else:
        return {"error": "not-git-repo"}


def build_remediation_info_url(
    base_url: str, version: Optional[str], spec: str, target_version: Optional[str] = ""
) -> Optional[str]:
    """
    Build the remediation info URL.

    Args:
        base_url (str): The base URL.
        version (Optional[str]): The current version.
        spec (str): The specification.
        target_version (Optional[str]): The target version.

    Returns:
        str: The remediation info URL.
    """

    params = {"from": version, "to": target_version}

    # No pinned version
    if not version:
        params = {"spec": spec}

    req = PreparedRequest()
    req.prepare_url(base_url, params)

    return req.url


def get_processed_options(
    policy_file: Dict[str, Any],
    ignore: Dict[str, Any],
    ignore_severity_rules: Dict[str, Any],
    exit_code: bool,
    ignore_unpinned_requirements: Optional[bool] = None,
    project: Optional[str] = None,
) -> Tuple[Dict[str, Any], Dict[str, Any], bool, Optional[bool], Optional[str]]:
    """
    Get processed options from the policy file.

    Args:
        policy_file (Dict[str, Any]): The policy file.
        ignore (Dict[str, Any]): The ignore settings.
        ignore_severity_rules (Dict[str, Any]): The ignore severity rules.
        exit_code (bool): The exit code setting.
        ignore_unpinned_requirements (Optional[bool]): The ignore unpinned requirements setting.
        project (Optional[str]): The project setting.

    Returns:
        Tuple[Dict[str, Any], Dict[str, Any], bool, Optional[bool], Optional[str]]: The processed options.
    """
    if policy_file:
        project_config = policy_file.get("project", {})
        security = policy_file.get("security", {})
        ctx = click.get_current_context()
        source = ctx.get_parameter_source("exit_code")

        if not project:
            project_id = project_config.get("id", None)
            if not project_id:
                project_id = None
            project = project_id

        if (
            ctx.get_parameter_source("ignore_unpinned_requirements")
            == click.core.ParameterSource.DEFAULT
        ):
            ignore_unpinned_requirements = security.get(
                "ignore-unpinned-requirements", None
            )

        if not ignore:
            ignore = security.get("ignore-vulnerabilities", {})
        if source == click.core.ParameterSource.DEFAULT:
            exit_code = not security.get("continue-on-vulnerability-error", False)
        ignore_cvss_below = security.get("ignore-cvss-severity-below", 0.0)
        ignore_cvss_unknown = security.get("ignore-cvss-unknown-severity", False)
        ignore_severity_rules = {
            "ignore-cvss-severity-below": ignore_cvss_below,
            "ignore-cvss-unknown-severity": ignore_cvss_unknown,
        }

    return (
        ignore,
        ignore_severity_rules,
        exit_code,
        ignore_unpinned_requirements,
        project,
    )


def get_fix_options(
    policy_file: Dict[str, Any], auto_remediation_limit: int
) -> Union[int, List[str]]:
    """
    Get fix options from the policy file.

    Args:
        policy_file (Dict[str, Any]): The policy file.
        auto_remediation_limit (int): The auto remediation limit.

    Returns:
        int: The auto remediation limit.
    """
    auto_fix = []

    source = click.get_current_context().get_parameter_source("auto_remediation_limit")
    if source == click.core.ParameterSource.COMMANDLINE:
        return auto_remediation_limit

    if policy_file:
        fix = policy_file.get("security-updates", {})
        auto_fix = fix.get("auto-security-updates-limit", None)
        if not auto_fix:
            auto_fix = []

    return auto_fix


class MutuallyExclusiveOption(click.Option):
    """
    A click option that is mutually exclusive with other options.
    """

    def __init__(self, *args, **kwargs):
        self.mutually_exclusive = set(kwargs.pop("mutually_exclusive", []))
        self.with_values = kwargs.pop("with_values", {})
        help = kwargs.get("help", "")
        if self.mutually_exclusive:
            ex_str = ", ".join(
                [
                    "{0} with values {1}".format(item, self.with_values.get(item))
                    if item in self.with_values
                    else item
                    for item in self.mutually_exclusive
                ]
            )
            kwargs["help"] = help + (
                " NOTE: This argument is mutually exclusive with "
                " arguments: [" + ex_str + "]."
            )
        super(MutuallyExclusiveOption, self).__init__(*args, **kwargs)

    def handle_parse_result(  # type: ignore
        self, ctx: click.Context, opts: Dict[str, Any], args: List[str]
    ) -> Tuple[Any, List[str]]:
        """
        Handle the parse result for mutually exclusive options.

        Args:
            ctx (click.Context): The click context.
            opts (Dict[str, Any]): The options dictionary.
            args (List[str]): The arguments list.

        Returns:
            Tuple[Any, List[str]]: The result and remaining arguments.
        """
        m_exclusive_used = self.mutually_exclusive.intersection(opts)
        option_used = m_exclusive_used and self.name in opts

        exclusive_value_used = False
        for used in m_exclusive_used:
            value_used = opts.get(used, None)
            if not isinstance(value_used, List):
                value_used = [value_used]
            if value_used and set(self.with_values.get(used, [])).intersection(
                value_used
            ):
                exclusive_value_used = True

        if option_used and (not self.with_values or exclusive_value_used):
            options = ", ".join(self.opts)
            prohibited = "".join(
                [
                    "\n * --{0} with {1}".format(item, self.with_values.get(item))
                    if item in self.with_values
                    else f"\n * {item}"
                    for item in self.mutually_exclusive
                ]
            )
            raise click.UsageError(
                f"Illegal usage: `{options}` is mutually exclusive with: {prohibited}"
            )

        return super(MutuallyExclusiveOption, self).handle_parse_result(ctx, opts, args)


class DependentOption(click.Option):
    """
    A click option that depends on other options.
    """

    def __init__(self, *args, **kwargs):
        self.required_options = set(kwargs.pop("required_options", []))
        help = kwargs.get("help", "")
        if self.required_options:
            ex_str = ", ".join(self.required_options)
            kwargs["help"] = help + (f" Requires: [ {ex_str} ]")
        super(DependentOption, self).__init__(*args, **kwargs)

    def handle_parse_result(  # type: ignore
        self, ctx: click.Context, opts: Dict[str, Any], args: List[str]
    ) -> Tuple[Any, List[str]]:
        """
        Handle the parse result for dependent options.

        Args:
            ctx (click.Context): The click context.
            opts (Dict[str, Any]): The options dictionary.
            args (List[str]): The arguments list.

        Returns:
            Tuple[Any, List[str]]: The result and remaining arguments.
        """
        missing_required_arguments = None

        if self.name in opts:
            missing_required_arguments = self.required_options.difference(opts)

        if missing_required_arguments:
            raise click.UsageError(
                "Illegal usage: `{}` needs the arguments `{}`.".format(
                    self.name, ", ".join(missing_required_arguments)
                )
            )

        return super(DependentOption, self).handle_parse_result(ctx, opts, args)


def transform_ignore(
    ctx: click.Context, param: click.Parameter, value: Tuple[str]
) -> Dict[str, Dict[str, Optional[str]]]:
    """
    Transform ignore parameters into a dictionary.

    Args:
        ctx (click.Context): The click context.
        param (click.Parameter): The click parameter.
        value (Tuple[str]): The parameter value.

    Returns:
        Dict[str, Dict[str, Optional[str]]]: The transformed ignore parameters.
    """
    ignored_default_dict = {"reason": "", "expires": None}
    if isinstance(value, tuple) and any(value):
        # Following code is required to support the 2 ways of providing 'ignore'
        # --ignore=1234,567,789
        # or, the historical way (supported for backward compatibility)
        # -i 1234 -i 567
        combined_value = ",".join(value)
        ignore_ids = {vuln_id.strip() for vuln_id in combined_value.split(",")}
        return {ignore_id: dict(ignored_default_dict) for ignore_id in ignore_ids}

    return {}


def active_color_if_needed(
    ctx: click.Context, param: click.Parameter, value: str
) -> str:
    """
    Activate color if needed based on the context and environment variables.

    Args:
        ctx (click.Context): The click context.
        param (click.Parameter): The click parameter.
        value (str): The parameter value.

    Returns:
        str: The parameter value.
    """
    if value == "screen":
        ctx.color = True

    color = os.environ.get("SAFETY_COLOR", None)

    if color is not None:
        color = color.lower()

        if color == "1" or color == "true":
            ctx.color = True
        elif color == "0" or color == "false":
            ctx.color = False

    return value


def json_alias(
    ctx: click.Context, param: click.Parameter, value: bool
) -> Optional[bool]:
    """
    Set the SAFETY_OUTPUT environment variable to 'json' if the parameter is used.

    Args:
        ctx (click.Context): The click context.
        param (click.Parameter): The click parameter.
        value (bool): The parameter value.

    Returns:
        bool: The parameter value.
    """
    if value:
        os.environ["SAFETY_OUTPUT"] = "json"
        return value


def html_alias(
    ctx: click.Context, param: click.Parameter, value: bool
) -> Optional[bool]:
    """
    Set the SAFETY_OUTPUT environment variable to 'html' if the parameter is used.

    Args:
        ctx (click.Context): The click context.
        param (click.Parameter): The click parameter.
        value (bool): The parameter value.

    Returns:
        bool: The parameter value.
    """
    if value:
        os.environ["SAFETY_OUTPUT"] = "html"
        return value


def bare_alias(
    ctx: click.Context, param: click.Parameter, value: bool
) -> Optional[bool]:
    """
    Set the SAFETY_OUTPUT environment variable to 'bare' if the parameter is used.

    Args:
        ctx (click.Context): The click context.
        param (click.Parameter): The click parameter.
        value (bool): The parameter value.

    Returns:
        bool: The parameter value.
    """
    if value:
        os.environ["SAFETY_OUTPUT"] = "bare"
        return value


def get_terminal_size() -> os.terminal_size:
    """
    Get the terminal size.

    Returns:
        os.terminal_size: The terminal size.
    """
    from shutil import get_terminal_size as t_size
    # get_terminal_size can report 0, 0 if run from pseudo-terminal prior Python 3.11 versions

    columns = t_size().columns or 80
    lines = t_size().lines or 24

    return os.terminal_size((columns, lines))


def clean_project_id(input_string: str) -> str:
    """
    Clean a project ID by removing non-alphanumeric characters and normalizing the string.

    Args:
        input_string (str): The input string.

    Returns:
        str: The cleaned project ID.
    """
    input_string = re.sub(r"[^a-zA-Z0-9]+", "-", input_string)
    input_string = input_string.strip("-")
    input_string = input_string.lower()

    return input_string


def validate_expiration_date(expiration_date: Optional[str]) -> Optional[datetime]:
    """
    Validate an expiration date string.

    Args:
        expiration_date (str): The expiration date string.

    Returns:
        Optional[datetime]: The validated expiration date if valid, None otherwise.
    """
    d = None

    if expiration_date:
        try:
            d = datetime.strptime(expiration_date, "%Y-%m-%d")
        except ValueError:
            pass

        try:
            d = datetime.strptime(expiration_date, "%Y-%m-%d %H:%M:%S")
        except ValueError:
            pass

    return d


class SafetyPolicyFile(click.ParamType):
    """
    Custom Safety Policy file to hold validations.
    """

    name = "filename"
    envvar_list_splitter = os.path.pathsep

    def __init__(
        self,
        mode: str = "r",
        encoding: Optional[str] = None,
        errors: str = "strict",
        pure: bool = os.environ.get("SAFETY_PURE_YAML", "false").lower() == "true",
    ) -> None:
        self.mode = mode
        self.encoding = encoding
        self.errors = errors
        self.basic_msg = "\n" + click.style(
            'Unable to load the Safety Policy file "{name}".', fg="red"
        )
        self.pure = pure

    def to_info_dict(self) -> Dict[str, Any]:
        """
        Convert the object to an info dictionary.

        Returns:
            Dict[str, Any]: The info dictionary.
        """
        info_dict = super().to_info_dict()
        info_dict.update(mode=self.mode, encoding=self.encoding)
        return info_dict

    def fail_if_unrecognized_keys(
        self,
        used_keys: List[str],
        valid_keys: List[str],
        param: Optional[click.Parameter] = None,
        ctx: Optional[click.Context] = None,
        msg: str = "{hint}",
        context_hint: str = "",
    ) -> None:
        """
        Fail if unrecognized keys are found in the policy file.

        Args:
            used_keys (List[str]): The used keys.
            valid_keys (List[str]): The valid keys.
            param (Optional[click.Parameter]): The click parameter.
            ctx (Optional[click.Context]): The click context.
            msg (str): The error message template.
            context_hint (str): The context hint for the error message.

        Raises:
            click.UsageError: If unrecognized keys are found.
        """
        for keyword in used_keys:
            if keyword not in valid_keys:
                match = None
                max_ratio = 0.0
                if isinstance(keyword, str):
                    for option in valid_keys:
                        ratio = SequenceMatcher(None, keyword, option).ratio()
                        if ratio > max_ratio:
                            match = option
                            max_ratio = ratio

                maybe_msg = (
                    f" Maybe you meant: {match}"
                    if max_ratio > 0.7
                    else f" Valid keywords in this level are: {', '.join(valid_keys)}"
                )

                self.fail(
                    msg.format(
                        hint=f'{context_hint}"{keyword}" is not a valid keyword.{maybe_msg}'
                    ),
                    param,
                    ctx,
                )

    def fail_if_wrong_bool_value(
        self, keyword: str, value: Any, msg: str = "{hint}"
    ) -> None:
        """
        Fail if a boolean value is invalid.

        Args:
            keyword (str): The keyword.
            value (Any): The value.
            msg (str): The error message template.

        Raises:
            click.UsageError: If the boolean value is invalid.
        """
        if value is not None and not isinstance(value, bool):
            self.fail(
                msg.format(
                    hint=f"'{keyword}' value needs to be a boolean. "
                    "You can use True, False, TRUE, FALSE, true or false"
                )
            )

    def convert(
        self, value: Any, param: Optional[click.Parameter], ctx: Optional[click.Context]
    ) -> Any:
        """
        Convert the parameter value to a Safety policy file.

        Args:
            value (Any): The parameter value.
            param (Optional[click.Parameter]): The click parameter.
            ctx (Optional[click.Context]): The click context.

        Returns:
            Any: The converted policy file.

        Raises:
            click.UsageError: If the policy file is invalid.
        """
        try:
            # Check if the value is already a file-like object
            if hasattr(value, "read") or hasattr(value, "write"):
                return value

            # Prepare the error message template
            msg = (
                self.basic_msg.format(name=value)
                + "\n"
                + click.style("HINT:", fg="yellow")
                + " {hint}"
            )

            # Open the file stream
            f, _ = click.types.open_stream(  # type: ignore
                value, self.mode, self.encoding, self.errors, atomic=False
            )
            filename = ""

            try:
                # Read the content of the file
                raw = f.read()
                yaml = YAML(typ="safe", pure=self.pure)
                safety_policy = yaml.load(raw)
                filename = f.name
                f.close()
            except Exception as e:
                # Handle YAML parsing errors
                show_parsed_hint = isinstance(e, MarkedYAMLError)
                hint = str(e)
                if show_parsed_hint:
                    hint = f"{str(e.problem).strip()} {str(e.context).strip()} {str(e.context_mark).strip()}"

                self.fail(msg.format(name=value, hint=hint), param, ctx)

            # Validate the structure of the safety policy
            if (
                not safety_policy
                or not isinstance(safety_policy, dict)
                or not safety_policy.get("security", None)
            ):
                hint = "you are missing the security root tag"
                try:
                    version = safety_policy["version"]
                    if version:
                        hint = (
                            f"{filename} is a policy file version {version}. "
                            "Legacy policy file parser only accepts versions minor than 3.0"
                            "\nNote: `safety check` command accepts policy file versions <= 2.0. Versions >= 2.0 are not supported."
                        )
                except Exception:
                    pass
                self.fail(msg.format(hint=hint), param, ctx)

            # Validate 'security' section keys
            security_config = safety_policy.get("security", {})
            security_keys = [
                "ignore-cvss-severity-below",
                "ignore-cvss-unknown-severity",
                "ignore-vulnerabilities",
                "continue-on-vulnerability-error",
                "ignore-unpinned-requirements",
            ]
            self.fail_if_unrecognized_keys(
                security_config.keys(),
                security_keys,
                param=param,
                ctx=ctx,
                msg=msg,
                context_hint='"security" -> ',
            )

            # Validate 'ignore-cvss-severity-below' value
            ignore_cvss_security_below = security_config.get(
                "ignore-cvss-severity-below", None
            )
            if ignore_cvss_security_below:
                limit = 0.0
                try:
                    limit = float(ignore_cvss_security_below)
                except ValueError:
                    self.fail(
                        msg.format(
                            hint="'ignore-cvss-severity-below' value needs to be an integer or float."
                        )
                    )
                if limit < 0 or limit > 10:
                    self.fail(
                        msg.format(
                            hint="'ignore-cvss-severity-below' needs to be a value between 0 and 10"
                        )
                    )

            # Validate 'continue-on-vulnerability-error' value
            continue_on_vulnerability_error = security_config.get(
                "continue-on-vulnerability-error", None
            )
            self.fail_if_wrong_bool_value(
                "continue-on-vulnerability-error", continue_on_vulnerability_error, msg
            )

            # Validate 'ignore-cvss-unknown-severity' value
            ignore_cvss_unknown_severity = security_config.get(
                "ignore-cvss-unknown-severity", None
            )
            self.fail_if_wrong_bool_value(
                "ignore-cvss-unknown-severity", ignore_cvss_unknown_severity, msg
            )

            # Validate 'ignore-vulnerabilities' section
            ignore_vulns = safety_policy.get("security", {}).get(
                "ignore-vulnerabilities", {}
            )
            if ignore_vulns:
                if not isinstance(ignore_vulns, dict):
                    self.fail(
                        msg.format(
                            hint="Vulnerability IDs under the 'ignore-vulnerabilities' key, need to "
                            "follow the convention 'ID_NUMBER:', probably you are missing a colon."
                        )
                    )

                normalized = {}

                for ignored_vuln_id, config in ignore_vulns.items():
                    ignored_vuln_config = config if config else {}

                    if not isinstance(ignored_vuln_config, dict):
                        self.fail(
                            msg.format(
                                hint=f"Wrong configuration under the vulnerability with ID: {ignored_vuln_id}"
                            )
                        )

                    context_msg = f'"security" -> "ignore-vulnerabilities" -> "{ignored_vuln_id}" -> '

                    self.fail_if_unrecognized_keys(
                        ignored_vuln_config.keys(),  # type: ignore
                        ["reason", "expires"],
                        param=param,
                        ctx=ctx,
                        msg=msg,
                        context_hint=context_msg,
                    )

                    reason = ignored_vuln_config.get("reason", "")
                    reason = str(reason) if reason else None
                    expires = ignored_vuln_config.get("expires", "")
                    expires = str(expires) if expires else None

                    try:
                        if int(ignored_vuln_id) < 0:
                            raise ValueError("Negative Vulnerability ID")
                    except ValueError:
                        self.fail(
                            msg.format(
                                hint=f"vulnerability id {ignored_vuln_id} under the 'ignore-vulnerabilities' root needs to "
                                f"be a positive integer"
                            )
                        )

                    # Validate expires date
                    d = validate_expiration_date(expires)

                    if expires and not d:
                        self.fail(
                            msg.format(
                                hint=f'{context_msg}expires: "{expires}" isn\'t a valid format '
                                f"for the expires keyword, "
                                "valid options are: YYYY-MM-DD or "
                                "YYYY-MM-DD HH:MM:SS"
                            )
                        )

                    normalized[str(ignored_vuln_id)] = {"reason": reason, "expires": d}

                safety_policy["security"]["ignore-vulnerabilities"] = normalized
                safety_policy["filename"] = filename
                safety_policy["raw"] = raw
            else:
                safety_policy["security"]["ignore-vulnerabilities"] = {}

            # Validate 'fix' section keys
            fix_config = safety_policy.get("fix", {})
            self.fail_if_unrecognized_keys(
                fix_config.keys(),
                ["auto-security-updates-limit"],
                param=param,
                ctx=ctx,
                msg=msg,
                context_hint='"fix" -> ',
            )
            auto_remediation_limit = fix_config.get("auto-security-updates-limit", None)

            if auto_remediation_limit:
                self.fail_if_unrecognized_keys(
                    auto_remediation_limit,
                    ["patch", "minor", "major"],
                    param=param,
                    ctx=ctx,
                    msg=msg,
                    context_hint='"auto-security-updates-limit" -> ',
                )

            return safety_policy
        except BadParameter as expected_e:
            raise expected_e
        except Exception as e:
            # Handle file not found errors gracefully, don't fail in the default case
            if ctx and isinstance(e, OSError):
                default = ctx.get_parameter_source
                source = (
                    default("policy_file")
                    if default("policy_file")
                    else default("policy_file_path")
                )
                if (
                    e.errno == 2
                    and source == click.core.ParameterSource.DEFAULT
                    and value == ".safety-policy.yml"
                ):
                    return None

            problem = click.style("Policy file YAML is not valid.")
            hint = click.style("HINT: ", fg="yellow") + str(e)
            self.fail(f"{problem}\n{hint}", param, ctx)

    def shell_complete(
        self, ctx: click.Context, param: click.Parameter, incomplete: str
    ):
        """
        Return a special completion marker that tells the completion
        system to use the shell to provide file path completions.

        Args:
            ctx (click.Context): The click context.
            param (click.Parameter): The click parameter.
            incomplete (str): The value being completed. May be empty.

        Returns:
            List[click.shell_completion.CompletionItem]: The completion items.

        .. versionadded:: 8.0
        """
        from click.shell_completion import CompletionItem

        return [CompletionItem(incomplete, type="file")]


class SingletonMeta(type):
    """
    A metaclass for singleton classes.
    """

    _instances: Dict[type, Any] = {}

    _lock: Lock = Lock()

    def __call__(cls, *args: Any, **kwargs: Any) -> Any:
        with cls._lock:
            if cls not in cls._instances:
                instance = super().__call__(*args, **kwargs)
                cls._instances[cls] = instance
        return cls._instances[cls]


class SafetyContext(metaclass=SingletonMeta):
    """
    A singleton class to hold the Safety context.
    """

    packages = []
    key = False
    db_mirror = False
    cached = None
    ignore_vulns = None
    ignore_severity_rules = None
    proxy = None
    include_ignored = False
    telemetry = None
    files = None
    stdin = None
    is_env_scan = None
    command: Optional[str] = None
    subcommand: Optional[str] = None
    review = None
    params = {}
    safety_source = "code"
    local_announcements = []
    scanned_full_path = []
    account = None


def sync_safety_context(f):
    """
    Decorator to sync the Safety context with the function arguments.
    """

    def new_func(*args, **kwargs):
        ctx = SafetyContext()

        legacy_key_added = False
        if "session" in kwargs:
            legacy_key_added = True
            session = kwargs.get("session")
            kwargs["key"] = session.api_key if session else None

        for attr in dir(ctx):
            if attr in kwargs:
                setattr(ctx, attr, kwargs.get(attr))

        if legacy_key_added:
            kwargs.pop("key")

        return f(*args, **kwargs)

    return new_func


@sync_safety_context
def get_packages_licenses(
    *,
    packages: Optional[List[Package]] = None,
    licenses_db: Optional[Dict[str, Any]] = None,
) -> List[Dict[str, Any]]:
    """
    Get the licenses for the specified packages based on their version.

    Args:
        packages (Optional[List[Package]]): The list of packages.
        licenses_db (Optional[Dict[str, Any]]): The licenses database.

    Returns:
        List[Dict[str, Any]]: The list of packages and their licenses.
    """
    SafetyContext().command = "license"

    if not packages:
        packages = []
    if not licenses_db:
        licenses_db = {}

    packages_licenses_db = licenses_db.get("packages", {})
    filtered_packages_licenses = []

    for pkg in packages:
        # Ignore recursive files not resolved
        if isinstance(pkg, RequirementFile):
            continue
        # normalize the package name
        pkg_name = canonicalize_name(pkg.name)
        # packages may have different licenses depending their version.
        pkg_licenses = packages_licenses_db.get(pkg_name, [])
        if not pkg.version:
            for req in pkg.requirements:
                if is_pinned_requirement(req.specifier):
                    pkg.version = next(iter(req.specifier)).version
                    break

            if not pkg.version:
                continue
        version_requested = parse_version(pkg.version)
        license_id = None
        license_name = None
        for pkg_version in pkg_licenses:
            license_start_version = parse_version(pkg_version["start_version"])
            # Stops and return the previous stored license when a new
            # license starts on a version above the requested one.
            if version_requested >= license_start_version:
                license_id = pkg_version["license_id"]
            else:
                # We found the license for the version requested
                break

        if license_id:
            license_name = get_license_name_by_id(license_id, licenses_db)
        if not license_id or not license_name:
            license_name = "unknown"

        filtered_packages_licenses.append(
            {"package": pkg_name, "version": pkg.version, "license": license_name}
        )

    return filtered_packages_licenses


def get_requirements_content(files: List[Any]) -> Dict[str, str]:
    """
    Get the content of the requirements files.

    Args:
        files (List[click.File]): The list of requirement files.

    Returns:
        Dict[str, str]: The content of the requirement files.

    Raises:
        InvalidProvidedReportError: If a file cannot be read.
    """
    requirements_files = {}

    for f in files:
        try:
            f.seek(0)
            requirements_files[f.name] = f.read()
            f.close()
        except Exception as e:
            raise InvalidProvidedReportError(
                message=f"Unable to read a requirement file scanned in the report. {e}"
            )

    return requirements_files


def is_ignore_unpinned_mode(version: str) -> bool:
    """
    Check if unpinned mode is enabled based on the version.

    Args:
        version (str): The version string.

    Returns:
        bool: True if unpinned mode is enabled, False otherwise.
    """
    ignore = SafetyContext().params.get("ignore_unpinned_requirements")
    return (ignore is None or ignore) and not version


def get_remediations_count(remediations: Dict[str, Any]) -> int:
    """
    Get the count of remediations.

    Args:
        remediations (Dict[str, Any]): The remediations dictionary.

    Returns:
        int: The count of remediations.
    """
    return sum((len(rem.keys()) for pkg, rem in remediations.items()))


def get_hashes(dependency: Any) -> List[Dict[str, str]]:
    """
    Get the hashes for a dependency.

    Args:
        dependency (Any): The dependency.

    Returns:
        List[Dict[str, str]]: The list of hashes.
    """
    pattern = re.compile(HASH_REGEX_GROUPS)

    return [
        {"method": method, "hash": hsh}
        for method, hsh in (
            pattern.match(d_hash).groups()  # type: ignore
            for d_hash in dependency.hashes
        )
    ]


def pluralize(word: str, count: int = 0) -> str:
    """
    Pluralize a word based on the count.

    Args:
        word (str): The word to pluralize.
        count (int): The count.

    Returns:
        str: The pluralized word.
    """
    if count == 1:
        return word

    default = {"was": "were", "this": "these", "has": "have"}

    if word in default:
        return default[word]

    if (
        word.endswith("s")
        or word.endswith("x")
        or word.endswith("z")
        or word.endswith("ch")
        or word.endswith("sh")
    ):
        return word + "es"

    if word.endswith("y"):
        if word[-2] in "aeiou":
            return word + "s"
        else:
            return word[:-1] + "ies"

    return word + "s"


def initialize_config_dirs() -> None:
    """
    Initialize the configuration directories.
    """
    USER_CONFIG_DIR.mkdir(parents=True, exist_ok=True)

    try:
        SYSTEM_CONFIG_DIR.mkdir(parents=True, exist_ok=True)
    except Exception:
        pass


def initialize_event_bus(ctx: Union["CustomContext", "typer.Context"]) -> bool:
    """
    Initializes the event bus for the given context. This should be called one
    time only per command run.
    The event bus requires the following conditions to be met:
    - Platform OR Platform and Firewall features enabled
    - Authenticated user
    Args:
        ctx (CustomContext): The context object containing necessary
                             information.
    Returns:
        bool: True if the event bus was successfully initialized,
              False otherwise.
    """
    try:
        obj: "SafetyCLI" = ctx.obj
        auth: Optional["Auth"] = None

        if obj and obj.events_enabled and (auth := getattr(obj, "auth", None)):
            client: "SafetyAuthSession" = auth.client
            token = client.token.get("access_token") if client.token else None

            # Start the event bus if the user has set up authn
            if client and bool(token or client.api_key):
                start_event_bus(obj, client)

                if event_bus := obj.event_bus:
                    # Trigger here CLI GROUP LOADED event
                    from safety.events.utils import (
                        create_internal_event,
                        InternalEventType,
                        InternalPayload,
                    )

                    event = create_internal_event(
                        event_type=InternalEventType.EVENT_BUS_READY,
                        payload=InternalPayload(ctx=ctx),
                    )

                    event_bus.emit(event)

                    return True

    except Exception as e:
        LOG.exception("Error starting event bus: %s", e)

    return False



import logging
from abc import ABCMeta, abstractmethod
from typing import Any, Dict, List, Tuple, Union, Optional

NOT_IMPLEMENTED = "You should implement this."

LOG = logging.getLogger(__name__)


class FormatterAPI:
    """
    Strategy Abstract class, with all the render methods that the concrete implementations should support.
    """

    __metaclass__ = ABCMeta

    def __init__(self, **kwargs: Any) -> None:
        """
        Dummy initializer for the FormatterAPI class.
        """
        pass

    @abstractmethod
    def render_vulnerabilities(self, announcements: List[Dict[str, Any]], vulnerabilities: List[Dict[str, Any]], remediations: Dict[str, Any], full: bool, packages: List[Dict[str, Any]], fixes: Tuple = ()) -> Optional[str]:
        """
        Render the vulnerabilities report.

        Args:
            announcements (List[Dict[str, Any]]): List of announcements.
            vulnerabilities (List[Dict[str, Any]]): List of vulnerabilities.
            remediations (Dict[str, Any]): Dictionary of remediations.
            full (bool): Whether to render a full report.
            packages (List[Dict[str, Any]]): List of packages.
            fixes (Tuple, optional): Tuple of fixes. Defaults to ().

        Returns:
            Optional[str]: Rendered vulnerabilities report.
        """
        raise NotImplementedError(NOT_IMPLEMENTED)  # pragma: no cover

    @abstractmethod
    def render_licenses(self, announcements: List[Dict[str, Any]], licenses: List[Dict[str, Any]]) -> Optional[str]:
        """
        Render the licenses report.

        Args:
            announcements (List[Dict[str, Any]]): List of announcements.
            licenses (List[Dict[str, Any]]): List of licenses.

        Returns:
            Optional[str]: Rendered licenses report.
        """
        raise NotImplementedError(NOT_IMPLEMENTED)  # pragma: no cover

    @abstractmethod
    def render_announcements(self, announcements: List[Dict[str, Any]]) -> Optional[str]:
        """
        Render the announcements.

        Args:
            announcements (List[Dict[str, Any]]): List of announcements.

        Returns:
            Optional[str]: Rendered announcements.
        """
        raise NotImplementedError(NOT_IMPLEMENTED)  # pragma: no cover


class SafetyFormatter(FormatterAPI):
    """
    Formatter class that implements the FormatterAPI to render reports in various formats.
    """
    def __init__(self, output: str, **kwargs: Any) -> None:
        """
        Initialize the SafetyFormatter with the specified output format.

        Args:
            output (str): The output format (e.g., 'json', 'html', 'bare', 'text').
            **kwargs: Additional keyword arguments.
        """
        from safety.formatters.screen import ScreenReport
        from safety.formatters.text import TextReport
        from safety.formatters.json import JsonReport
        from safety.formatters.bare import BareReport
        from safety.formatters.html import HTMLReport

        self.format = ScreenReport(**kwargs)

        if output == 'json':
            self.format = JsonReport(**kwargs)
        elif output == 'html':
            self.format = HTMLReport(**kwargs)
        elif output == 'bare':
            self.format = BareReport(**kwargs)
        elif output == 'text':
            self.format = TextReport(**kwargs)

    def render_vulnerabilities(self, announcements: List[Dict[str, Any]], vulnerabilities: List[Dict[str, Any]], remediations: Dict[str, Any], full: bool, packages: List[Dict[str, Any]], fixes: Tuple = ()) -> Optional[str]:
        """
        Render the vulnerabilities report.

        Args:
            announcements (List[Dict[str, Any]]): List of announcements.
            vulnerabilities (List[Dict[str, Any]]): List of vulnerabilities.
            remediations (Dict[str, Any]): Dictionary of remediations.
            full (bool): Whether to render a full report.
            packages (List[Dict[str, Any]]): List of packages.
            fixes (Tuple, optional): Tuple of fixes. Defaults to ().

        Returns:
            Optional[str]: Rendered vulnerabilities report.
        """
        LOG.info('Safety is going to render_vulnerabilities with format: %s', self.format)
        return self.format.render_vulnerabilities(announcements, vulnerabilities, remediations, full, packages, fixes)

    def render_licenses(self, announcements: List[Dict[str, Any]], licenses: List[Dict[str, Any]]) -> Optional[str]:
        """
        Render the licenses report.

        Args:
            announcements (List[Dict[str, Any]]): List of announcements.
            licenses (List[Dict[str, Any]]): List of licenses.

        Returns:
            Optional[str]: Rendered licenses report.
        """
        LOG.info('Safety is going to render_licenses with format: %s', self.format)
        return self.format.render_licenses(announcements, licenses)

    def render_announcements(self, announcements: List[Dict[str, Any]]):
        """
        Render the announcements.

        Args:
            announcements (List[Dict[str, Any]]): List of announcements.

        Returns:
            Optional[str]: Rendered announcements.
        """
        LOG.info('Safety is going to render_announcements with format: %s', self.format)
        return self.format.render_announcements(announcements)



from functools import wraps
from typing import TYPE_CHECKING, Any, Callable, List, Optional, TypeVar, cast, overload


if TYPE_CHECKING:
    from safety.events.event_bus import EventBus
    from safety.cli_util import CustomContext


def should_emit(
    event_bus: Optional["EventBus"], ctx: Optional["CustomContext"]
) -> bool:
    """
    Common conditions that apply to all event emissions.
    """
    if event_bus is None:
        return False

    # Be aware that ctx depends on the command being parsed, if the emit func
    # is called from the entrypoint group command, ctx will not have
    # the command parsed yet.

    return True


def should_emit_firewall_heartbeat(ctx: Optional["CustomContext"]) -> bool:
    """
    Condition to check if the firewall is enabled.
    """
    if ctx and ctx.obj.firewall_enabled:
        return True

    return False


# Define TypeVars for better typing
F = TypeVar("F", bound=Callable[..., Any])
R = TypeVar("R")


@overload
def conditional_emitter(emit_func: F, *, conditions: None = None) -> F: ...


@overload
def conditional_emitter(
    emit_func: None = None,
    *,
    conditions: Optional[List[Callable[[Optional["CustomContext"]], bool]]] = None,
) -> Callable[[F], F]: ...


def conditional_emitter(
    emit_func=None,
    *,
    conditions: Optional[List[Callable[[Optional["CustomContext"]], bool]]] = None,
):
    """
    A decorator that conditionally calls the decorated function based on conditions.
    Only executes the decorated function if all conditions evaluate to True.
    """

    def decorator(func: F) -> F:
        @wraps(func)
        def wrapper(event_bus, ctx=None, *args, **kwargs):
            if not should_emit(event_bus, ctx):
                return None

            if conditions:
                if all(condition(ctx) for condition in conditions):
                    return func(event_bus, ctx, *args, **kwargs)
                return None
            return func(event_bus, ctx, *args, **kwargs)

        return cast(F, wrapper)  # Cast to help type checker

    if emit_func is None:
        return decorator
    return decorator(emit_func)



from typing import TYPE_CHECKING
from .bus import EventBus

from safety_schemas.models.events import EventType
from safety.events.types import InternalEventType
from safety.events.handlers import SecurityEventsHandler

from safety.constants import PLATFORM_API_EVENTS_ENDPOINT

if TYPE_CHECKING:
    from safety.models import SafetyCLI
    from safety.auth.utils import SafetyAuthSession


def start_event_bus(obj: "SafetyCLI", session: "SafetyAuthSession"):
    """
    Initializes the event bus with the default security events handler
    for authenticated users.
    This function creates an instance of the EventBus, starts it,
    and assigns it to the `event_bus` attribute of the provided `obj`.
    It also initializes the `security_events_handler` with the necessary
    parameters and subscribes it to a predefined list of events.

    Args:
        obj (SafetyCLI): The main application object.
        session (SafetyAuthSession): The authentication session containing
                                     the necessary credentials and proxies.

    """
    event_bus = EventBus()
    event_bus.start()
    obj.event_bus = event_bus

    obj.security_events_handler = SecurityEventsHandler(
        api_endpoint=PLATFORM_API_EVENTS_ENDPOINT,
        proxies=session.proxies,  # type: ignore
        auth_token=session.token.get("access_token"),
        api_key=session.api_key,
    )

    events = [
        EventType.COMMAND_EXECUTED,
        EventType.COMMAND_ERROR,
        InternalEventType.CLOSE_RESOURCES,
        InternalEventType.FLUSH_SECURITY_TRACES,
    ]

    event_bus.subscribe(events, obj.security_events_handler)

    if obj.firewall_enabled:
        from safety.firewall.events.utils import register_event_handlers

        register_event_handlers(obj.event_bus, obj=obj)



import random
import time
import os
from rich.text import Text
from rich.style import Style
from safety.scan.fun_mode.celebration_effects import CELEBRATION_EFFECTS

# -----------------------------
# Data: ASCII Arts, Fortunes, EMOJIS, etc.
# -----------------------------

ASCII_ARTS = {
    "ascii": [
        # Hedgehog guarding a shield
        r"""
        /\     /\
       {  `---'  }
       {  O   O  }
       ~~>  V  <~~
        \  \|/  /
         `-----'__
         /     \  `^\_
        {       }\ |\_\_   *Safely Protected from vulnerabilities!*
        |  \_/  |/ /  \_\_
         \__/  /(_E     \__/
           (  /
            MM
        """,

        # Cat with a shield
        r"""
       /\_/\
     =( °w° )=   *Purrr... no vulnerabilities dare to cross!*
      (   *   )
     ---(   )---
        /     \
       /  ^    \
      ( (   )   )
       \_) |_(_/
      __||___||__
     |           |
     |   SAFE    |
     |    &      |
     |  SECURE!  |
     |___________|
        /     \
       |       |

        """,

        # Bunny with a shield
        r"""
     (\_/)
     ( •_•)  *Hop-hop! SafetyCLI ready, no vulns here!*
    ( >🥕< )
     /     \
    / |   | \
   /  |   |  \
  /   |___|   \
 (               )
  \____|||_____/
       ||||
      __||__
     |      |
     | SAFE |
     | FROM |
     | BUGS |
     |______|

        """,

        # Dog behind a shield
        r"""
         / \__
        (    o\____
        /         O
       /   (_______/    *Woof! Our shield is strong, no vulns inside!*
      /_____/


        """
    ]
}

FORTUNES = [
    "Your dependencies are safer than a password-manager's vault.",
    "Your code sparkles with zero known vulnerabilities!",
    "All vulnerabilities fear your security prowess!",
    "Your build is as solid as a rock!",
    "Your code is a fortress; no bug can breach it.",
    "In the realm of code, you are the vigilant guardian.",
    "Each line you write fortifies the castle of your code.",
    "Your code is a well-oiled machine, impervious to rust.",
    "Your code is a symphony of security and efficiency.",
    "Your code is a beacon of safety in the digital ocean.",
    "Your code is a masterpiece, untouched by the hands of vulnerabilities.",
    "Your code stands tall, a citadel against cyber threats.",
    "Your code is a tapestry woven with threads of safety.",
    "Your code is a lighthouse, guiding ships away from the rocks of vulnerabilities.",
    "Your code is a garden where no weeds of bugs can grow.",
    "In the realm of software, your security measures are legendary.",
]


EMOJIS = [
        "🐶", "🐕", "🦮", "🐕‍🦺", "🐩", "🐺", "🦊",
        "🐱", "🐈", "🐈‍⬛", "🦁", "🐯", "🐅", "🐆", "🐴", "🫎", "🫏",
        "🐎", "🦄", "🦓", "🦌", "🦬", "🐮", "🐂", "🐃", "🐄",
        "🐏", "🐑", "🐐", "🐪", "🐫", "🦙", "🦒",
        "🐘", "🦣", "🦏", "🦛", "🐭", "🐁", "🐀", "🐹", "🐰", "🐇",
        "🐿️", "🦫", "🦔", "🦇", "🐻", "🐻‍❄️", "🐨", "🐼", "🦦",
        "🦘", "🦡", "🦃", "🐔", "🐓", "🐤", "🐥", "🐦", "🐧",
        "🕊️", "🦅", "🦆", "🦢", "🦉", "🦤", "🦩", "🦚", "🦜", "🐦‍⬛",
        "🪿", "🐦‍🔥", "🐸", "🐊", "🐢", "🦎", "🐉", "🦕",
        "🦖", "🐳", "🐋", "🐬", "🦭", "🐟", "🐠", "🐡", "🦈", "🐙",
        "🪼", "🦀", "🦞", "🦑", "🐌", "🦋", "🐛","🐝"
    ]

# -----------------------------
# Helper functions (Effects)
# -----------------------------

def show_race(console):
    # Pick two different EMOJIS at random
    emoji1, emoji2 = random.sample(EMOJIS, 2)
    finish_line = 50
    pos1 = 0
    pos2 = 0

    console.print("Ready... Set... Go!", style="bold cyan")
    time.sleep(1)
    console.clear()

    while True:
        # Move contestants forward by random increments
        pos1 += random.randint(1, 3)
        pos2 += random.randint(1, 3)

        console.clear()
        console.print("[green]Finish line[/green]" + " " * (finish_line - 10) + "|")
        line1 = " " * pos1 + emoji1
        line2 = " " * pos2 + emoji2
        console.print(f"{emoji1} lane:  {line1}")
        console.print(f"{emoji2} lane:  {line2}")

        time.sleep(0.3)

        finished1 = pos1 >= finish_line
        finished2 = pos2 >= finish_line

        if finished1 and finished2:
            console.print("It's a tie! Both reached the finish line at the same time!", style="bold magenta")
            break
        elif finished1:
            console.print(f"The {emoji1} wins! Slow and steady (or maybe fast?), it prevailed!", style="bold green")
            break
        elif finished2:
            console.print(f"The {emoji2} wins! Speed and agility triumphed!", style="bold green")
            break

    time.sleep(2)
    console.clear()
    console.print("Hope you enjoyed the race! 🎉", style="bold cyan")


# -----------------------------
# Main Easter Egg Dispatcher
# -----------------------------

def run_easter_egg(console, exit_code: int) -> None:
    """
    Runs an easter egg based on the SAFETY_FUN_MODE environment variable.
    This function can be easily removed or commented out.
    """
    egg_mode = os.getenv("SAFETY_FUN_MODE", "").strip().lower()

    allowed_modes = {"ascii", "fx", "race", "fortune"}

    if exit_code == 0 and egg_mode in allowed_modes:
        if egg_mode == "ascii":
            art = random.choice(ASCII_ARTS["ascii"])
            console.print(art, style="green")

        elif egg_mode == "fx":
            effect = random.choice(CELEBRATION_EFFECTS)
            effect(console)  # Run the randomly selected effect

        elif egg_mode == "race":
            show_race(console)

        elif egg_mode == "fortune":
            fortune_message = random.choice(FORTUNES)
            console.print(f"\n[italic cyan]{fortune_message}[/italic cyan]\n")



import itertools
import logging
import re
import sys
from typing import Any, Optional

import click

try:
    import github as pygithub
except ImportError:
    pygithub = None

from packaging.specifiers import SpecifierSet
from packaging.utils import canonicalize_name

from . import utils, requirements

LOG = logging.getLogger(__name__)


def create_branch(repo: Any, base_branch: str, new_branch: str) -> None:
    """
    Create a new branch in the given GitHub repository.

    Args:
        repo (Any): The GitHub repository object.
        base_branch (str): The name of the base branch.
        new_branch (str): The name of the new branch to create.
    """
    ref = repo.get_git_ref("heads/" + base_branch)
    repo.create_git_ref(ref="refs/heads/" + new_branch, sha=ref.object.sha)


def delete_branch(repo: Any, branch: str) -> None:
    """
    Delete a branch from the given GitHub repository.

    Args:
        repo (Any): The GitHub repository object.
        branch (str): The name of the branch to delete.
    """
    ref = repo.get_git_ref(f"heads/{branch}")
    ref.delete()

@click.command()
@click.option('--repo', help='GitHub standard repo path (eg, my-org/my-project)')
@click.option('--token', help='GitHub Access Token')
@click.option('--base-url', help='Optional custom Base URL, if you\'re using GitHub enterprise', default=None)
@click.pass_obj
@utils.require_files_report
def github_pr(obj: Any, repo: str, token: str, base_url: Optional[str]) -> None:
    """
    Create a GitHub PR to fix any vulnerabilities using Safety's remediation data.

    This is usually run by a GitHub action. If you're running this manually, ensure that your local repo is up to date and on HEAD - otherwise you'll see strange results.

    Args:
        obj (Any): The Click context object containing report data.
        repo (str): The GitHub repository path.
        token (str): The GitHub Access Token.
        base_url (Optional[str]): Custom base URL for GitHub Enterprise, if applicable.
    """
    if pygithub is None:
        click.secho("pygithub is not installed. Did you install Safety with GitHub support? Try pip install safety[github]", fg='red')
        sys.exit(1)

    # Load alert configurations from the policy
    alert = obj.policy.get('alert', {}) or {}
    security = alert.get('security', {}) or {}
    config_pr = security.get('github-pr', {}) or {}

    branch_prefix = config_pr.get('branch-prefix', 'pyup/')
    pr_prefix = config_pr.get('pr-prefix', '[PyUp] ')
    assignees = config_pr.get('assignees', [])
    labels = config_pr.get('labels', ['security'])
    label_severity = config_pr.get('label-severity', True)
    ignore_cvss_severity_below = config_pr.get('ignore-cvss-severity-below', 0)
    ignore_cvss_unknown_severity = config_pr.get('ignore-cvss-unknown-severity', False)

    # Authenticate with GitHub
    gh = pygithub.Github(token, **({"base_url": base_url} if base_url else {}))
    repo_name = repo
    repo = gh.get_repo(repo)
    try:
        self_user = gh.get_user().login
    except pygithub.GithubException:
        # If we're using a token from an action (or integration) we can't call `get_user()`. Fall back
        # to assuming we're running under an action
        self_user = "web-flow"

    # Collect all remediations from the report
    req_remediations = list(itertools.chain.from_iterable(
        rem.get('requirements', {}).values() for pkg_name, rem in obj.report['remediations'].items()))

    # Get all open pull requests for the repository
    pulls = repo.get_pulls(state='open', sort='created', base=repo.default_branch)
    pending_updates = set(
        [f"{canonicalize_name(req_rem['requirement']['name'])}{req_rem['requirement']['specifier']}" for req_rem in req_remediations])

    created = 0

    # TODO: Refactor this loop into a fn to iterate over remediations nicely
    # Iterate over all requirements files and process each remediation
    for name, contents in obj.requirements_files.items():
        raw_contents = contents
        contents = contents.decode('utf-8') # TODO - encoding?
        parsed_req_file = requirements.RequirementFile(name, contents)

        for remediation in req_remediations:
            pkg = remediation['requirement']['name']
            pkg_canonical_name: str = canonicalize_name(pkg)
            analyzed_spec: str = remediation['requirement']['specifier']

            # Skip remediations without a recommended version
            if remediation['recommended_version'] is None:
                LOG.debug(f"The GitHub PR alerter only currently supports remediations that have a recommended_version: {pkg}")
                continue

            # We have a single remediation that can have multiple vulnerabilities
            # Find all vulnerabilities associated with the remediation
            vulns = [x for x in obj.report['vulnerabilities'] if
                     x['package_name'] == pkg_canonical_name and x['analyzed_requirement']['specifier'] == analyzed_spec]

            # Skip if all vulnerabilities have unknown severity and the ignore flag is set
            if ignore_cvss_unknown_severity and all(x['severity'] is None for x in vulns):
                LOG.debug("All vulnerabilities have unknown severity, and ignore_cvss_unknown_severity is set.")
                continue

            highest_base_score = 0
            for vuln in vulns:
                if vuln['severity'] is not None:
                    highest_base_score = max(highest_base_score, (vuln['severity'].get('cvssv3', {}) or {}).get('base_score', 10))

            # Skip if none of the vulnerabilities meet the severity threshold
            if ignore_cvss_severity_below:
                at_least_one_match = False
                for vuln in vulns:
                    # Consider a None severity as a match, since it's controlled by a different flag
                    # If we can't find a base_score but we have severity data, assume it's critical for now.
                    if vuln['severity'] is None or (vuln['severity'].get('cvssv3', {}) or {}).get('base_score', 10) >= ignore_cvss_severity_below:
                        at_least_one_match = True

                if not at_least_one_match:
                    LOG.debug(f"None of the vulnerabilities found have a score greater than or equal to the ignore_cvss_severity_below of {ignore_cvss_severity_below}")
                    continue

            for parsed_req in parsed_req_file.requirements:
                specs = SpecifierSet('>=0') if parsed_req.specs == SpecifierSet('') else parsed_req.specs

                # Check if the requirement matches the remediation
                if canonicalize_name(parsed_req.name) == pkg_canonical_name and str(specs) == analyzed_spec:
                    updated_contents = parsed_req.update_version(contents, remediation['recommended_version'])
                    pending_updates.discard(f"{pkg_canonical_name}{analyzed_spec}")

                    new_branch = branch_prefix + utils.generate_branch_name(pkg, remediation)
                    skip_create = False

                    # Few possible cases:
                    # 1. No existing PRs exist for this change (don't need to handle)
                    # 2. An existing PR exists, and it's out of date (eg, recommended 0.5.1 and we want 0.5.2)
                    # 3. An existing PR exists, and it's not mergable anymore (eg, needs a rebase)
                    # 4. An existing PR exists, and everything's up to date.
                    # 5. An existing PR exists, but it's not needed anymore (perhaps we've been updated to a later version)
                    # 6. No existing PRs exist, but a branch does exist (perhaps the PR was closed but a stale branch left behind)
                    # In any case, we only act if we've been the only committer to the branch.
                    # Handle various cases for existing pull requests
                    for pr in pulls:
                        if not pr.head.ref.startswith(branch_prefix):
                            continue

                        authors = [commit.committer.login for commit in pr.get_commits()]
                        only_us = all([x == self_user for x in authors])

                        try:
                            _, pr_pkg, pr_spec, pr_ver = pr.head.ref.split('/')
                        except ValueError:
                            # It's possible that something weird has manually been done, so skip that
                            # Skip invalid branch names
                            LOG.debug('Found an invalid branch name on an open PR, that matches our prefix. Skipping.')
                            continue

                        pr_pkg = canonicalize_name(pr_pkg)

                        if pr_pkg != pkg_canonical_name:
                            continue

                        # Case 4: An up-to-date PR exists
                        if pr_pkg == pkg_canonical_name and pr_spec == analyzed_spec and pr_ver == \
                                remediation['recommended_version'] and pr.mergeable:
                            LOG.debug(f"An up to date PR #{pr.number} for {pkg} was found, no action will be taken.")

                            skip_create = True
                            continue

                        if not only_us:
                            LOG.debug(f"There are other committers on the PR #{pr.number} for {pkg}. No further action will be taken.")
                            continue

                        # Case 2: An existing PR is out of date
                        if pr_pkg == pkg_canonical_name and pr_spec == analyzed_spec and pr_ver != \
                                remediation['recommended_version']:
                            LOG.debug(f"Closing stale PR #{pr.number} for {pkg} as a newer recommended version became")

                            pr.create_issue_comment("This PR has been replaced, since a newer recommended version became available.")
                            pr.edit(state='closed')
                            delete_branch(repo, pr.head.ref)

                        # Case 3: An existing PR is not mergeable
                        if not pr.mergeable:
                            LOG.debug(f"Closing PR #{pr.number} for {pkg} as it has become unmergable and we were the only committer")

                            pr.create_issue_comment("This PR has been replaced since it became unmergable.")
                            pr.edit(state='closed')
                            delete_branch(repo, pr.head.ref)

                    # Skip if no changes were made
                    if updated_contents == contents:
                        LOG.debug(f"Couldn't update {pkg} to {remediation['recommended_version']}")
                        continue

                    # Skip creation if indicated
                    if skip_create:
                        continue

                    # Create a new branch and commit the changes
                    try:
                        create_branch(repo, repo.default_branch, new_branch)
                    except pygithub.GithubException as e:
                        if e.data['message'] == "Reference already exists":
                            # There might be a stale branch. If the bot is the only committer, nuke it.
                            comparison = repo.compare(repo.default_branch, new_branch)
                            authors = [commit.committer.login for commit in comparison.commits]
                            only_us = all([x == self_user for x in authors])

                            if only_us:
                                delete_branch(repo, new_branch)
                                create_branch(repo, repo.default_branch, new_branch)
                            else:
                                LOG.debug(f"The branch '{new_branch}' already exists - but there is no matching PR and this branch has committers other than us. This remediation will be skipped.")
                                continue
                        else:
                            raise e

                    try:
                        repo.update_file(
                                path=name,
                                message=utils.generate_commit_message(pkg, remediation),
                                content=updated_contents,
                                branch=new_branch,
                                sha=utils.git_sha1(raw_contents)
                            )
                    except pygithub.GithubException as e:
                        if "does not match" in e.data['message']:
                            click.secho(f"GitHub blocked a commit on our branch to the requirements file, {name}, as the local hash we computed didn't match the version on {repo.default_branch}. Make sure you're running safety against the latest code on your default branch.", fg='red')
                            continue
                        else:
                            raise e

                    pr = repo.create_pull(title=pr_prefix + utils.generate_title(pkg, remediation, vulns), body=utils.generate_body(pkg, remediation, vulns, api_key=obj.key), head=new_branch, base=repo.default_branch)
                    LOG.debug(f"Created Pull Request to update {pkg}")

                    created += 1

                    # Add assignees and labels to the PR
                    for assignee in assignees:
                        pr.add_to_assignees(assignee)

                    for label in labels:
                        pr.add_to_labels(label)

                    if label_severity:
                        score_as_label = utils.cvss3_score_to_label(highest_base_score)
                        if score_as_label:
                            pr.add_to_labels(score_as_label)

    if len(pending_updates) > 0:
        click.secho("The following remediations were not followed: {}".format(', '.join(pending_updates)), fg='red')

    if created:
        click.secho(f'Safety successfully created {created} GitHub PR{"s" if created > 1 else ""} for repo {repo_name}')
    else:
        click.secho('No PRs created; please run the command with debug mode for more information.')


@click.command()
@click.option('--repo', help='GitHub standard repo path (eg, my-org/my-project)')
@click.option('--token', help='GitHub Access Token')
@click.option('--base-url', help='Optional custom Base URL, if you\'re using GitHub enterprise', default=None)
@click.pass_obj
@utils.require_files_report # TODO: For now, it can be removed in the future to support env scans.
def github_issue(obj: Any, repo: str, token: str, base_url: Optional[str]) -> None:
    """
    Create a GitHub Issue for any vulnerabilities found using PyUp's remediation data.

    Normally, this is run by a GitHub action. If you're running this manually, ensure that your local repo is up to date and on HEAD - otherwise you'll see strange results.

    Args:
        obj (Any): The Click context object containing report data.
        repo (str): The GitHub repository path.
        token (str): The GitHub Access Token.
        base_url (Optional[str]): Custom base URL for GitHub Enterprise, if applicable.
    """
    LOG.info(f'github_issue')

    if pygithub is None:
        click.secho("pygithub is not installed. Did you install Safety with GitHub support? Try pip install safety[github]", fg='red')
        sys.exit(1)

    # Load alert configurations from the policy
    alert = obj.policy.get('alert', {}) or {}
    security = alert.get('security', {}) or {}
    config_issue = security.get('github-issue', {}) or {}

    issue_prefix = config_issue.get('issue-prefix', '[PyUp] ')
    assignees = config_issue.get('assignees', [])
    labels = config_issue.get('labels', ['security'])

    label_severity = config_issue.get('label-severity', True)
    ignore_cvss_severity_below = config_issue.get('ignore-cvss-severity-below', 0)
    ignore_cvss_unknown_severity = config_issue.get('ignore-cvss-unknown-severity', False)

    # Authenticate with GitHub
    gh = pygithub.Github(token, **({"base_url": base_url} if base_url else {}))
    repo_name = repo
    repo = gh.get_repo(repo)

    # Get all open issues for the repository
    issues = list(repo.get_issues(state='open', sort='created'))
    ISSUE_TITLE_REGEX = re.escape(issue_prefix) + r"Security Vulnerability in (.+)"
    req_remediations = list(itertools.chain.from_iterable(
        rem.get('requirements', {}).values() for pkg_name, rem in obj.report['remediations'].items()))

    created = 0

    # Iterate over all requirements files and process each remediation
    for name, contents in obj.requirements_files.items():
        raw_contents = contents
        contents = contents.decode('utf-8') # TODO - encoding?
        parsed_req_file = requirements.RequirementFile(name, contents)

        for remediation in req_remediations:
            pkg: str = remediation['requirement']['name']
            pkg_canonical_name: str = canonicalize_name(pkg)
            analyzed_spec: str = remediation['requirement']['specifier']

            # Skip remediations without a recommended version
            if remediation['recommended_version'] is None:
                LOG.debug(f"The GitHub Issue alerter only currently supports remediations that have a recommended_version: {pkg}")
                continue

            # We have a single remediation that can have multiple vulnerabilities
            # Find all vulnerabilities associated with the remediation
            vulns = [x for x in obj.report['vulnerabilities'] if x['package_name'] == pkg_canonical_name and x['analyzed_requirement']['specifier'] == analyzed_spec]

            # Skip if all vulnerabilities have unknown severity and the ignore flag is set
            if ignore_cvss_unknown_severity and all(x['severity'] is None for x in vulns):
                LOG.debug("All vulnerabilities have unknown severity, and ignore_cvss_unknown_severity is set.")
                continue

            highest_base_score = 0
            for vuln in vulns:
                if vuln['severity'] is not None:
                    highest_base_score = max(highest_base_score, (vuln['severity'].get('cvssv3', {}) or {}).get('base_score', 10))

            # Skip if none of the vulnerabilities meet the severity threshold
            if ignore_cvss_severity_below:
                at_least_one_match = False
                for vuln in vulns:
                    # Consider a None severity as a match, since it's controlled by a different flag
                    # If we can't find a base_score but we have severity data, assume it's critical for now.
                    if vuln['severity'] is None or (vuln['severity'].get('cvssv3', {}) or {}).get('base_score', 10) >= ignore_cvss_severity_below:
                        at_least_one_match = True
                        break

                if not at_least_one_match:
                    LOG.debug(f"None of the vulnerabilities found have a score greater than or equal to the ignore_cvss_severity_below of {ignore_cvss_severity_below}")
                    continue

            for parsed_req in parsed_req_file.requirements:
                specs = SpecifierSet('>=0') if parsed_req.specs == SpecifierSet('') else parsed_req.specs
                if canonicalize_name(parsed_req.name) == pkg_canonical_name and str(specs) == analyzed_spec:
                    skip = False
                    for issue in issues:
                        match = re.match(ISSUE_TITLE_REGEX, issue.title)
                        if match:
                            group = match.group(1)
                            if group == f"{pkg}{analyzed_spec}" or group == f"{pkg_canonical_name}{analyzed_spec}":
                                skip = True
                                break

                    # For now, we just skip issues if they already exist - we don't try and update them.
                    # Skip if an issue already exists for this remediation
                    if skip:
                        LOG.debug(
                            f"An issue already exists for {pkg}{analyzed_spec} - skipping")
                        continue

                    # Create a new GitHub issue
                    pr = repo.create_issue(title=issue_prefix + utils.generate_issue_title(pkg, remediation), body=utils.generate_issue_body(pkg, remediation, vulns, api_key=obj.key))
                    created += 1
                    LOG.debug(f"Created issue to update {pkg}")

                    # Add assignees and labels to the issue
                    for assignee in assignees:
                        pr.add_to_assignees(assignee)

                    for label in labels:
                        pr.add_to_labels(label)

                    if label_severity:
                        score_as_label = utils.cvss3_score_to_label(highest_base_score)
                        if score_as_label:
                            pr.add_to_labels(score_as_label)

    if created:
        click.secho(f'Safety successfully created {created} new GitHub Issue{"s" if created > 1 else ""} for repo {repo_name}')
    else:
        click.secho('No issues created; please run the command with debug mode for more information.')






from pathlib import Path
from safety_schemas.models import Ecosystem, FileType
from typer import FileTextWrite

from .python.main import PythonFile


class InspectableFileContext:
    """
    Context manager for handling the lifecycle of an inspectable file.

    This class ensures that the file is properly opened and closed, handling any
    exceptions that may occur during the process.
    """

    def __init__(self, file_path: Path,
                 file_type: FileType) -> None:
        """
        Initializes the InspectableFileContext.

        Args:
            file_path (Path): The path to the file.
            file_type (FileType): The type of the file.
        """
        self.file_path = file_path
        self.inspectable_file = None
        self.file_type = file_type

    def __enter__(self): # TODO: Handle permission issue /Applications/...
        """
        Enters the runtime context related to this object.

        Opens the file and creates the appropriate inspectable file object based on the file type.

        Returns:
            The inspectable file object.
        """
        try:
            file: FileTextWrite = open(self.file_path, mode='r+') # type: ignore
            self.inspectable_file = TargetFile.create(file_type=self.file_type, file=file)
        except Exception as e:
            # TODO: Report this
            pass

        return self.inspectable_file

    def __exit__(self, exc_type, exc_value, traceback):
        """
        Exits the runtime context related to this object.

        Ensures that the file is properly closed.
        """
        if self.inspectable_file:
            self.inspectable_file.file.close()

class TargetFile():
    """
    Factory class for creating inspectable file objects based on the file type and ecosystem.
    """

    @classmethod
    def create(cls, file_type: FileType, file: FileTextWrite):
        """
        Creates an inspectable file object based on the file type and ecosystem.

        Args:
            file_type (FileType): The type of the file.
            file (FileTextWrite): The file object.

        Returns:
            An instance of the appropriate inspectable file class.

        Raises:
            ValueError: If the ecosystem or file type is unsupported.
        """
        if file_type.ecosystem == Ecosystem.PYTHON:
            return PythonFile(file=file, file_type=file_type)

        raise ValueError("Unsupported ecosystem or file type: " \
                         f"{file_type.ecosystem}:{file_type.value}")



import logging
import os
import shutil
from pathlib import Path
from sys import platform
from typing import TYPE_CHECKING, List

from .base import CommandInterceptor, Tool
from .types import InterceptorType

if TYPE_CHECKING or platform == "win32":
    import winreg

LOG = logging.getLogger(__name__)


class WindowsInterceptor(CommandInterceptor):
    def __init__(self):
        super().__init__(InterceptorType.WINDOWS_BAT)
        self.scripts_dir = Path.home() / "AppData" / "Local" / "safety"
        self.backup_dir = self.scripts_dir / "backups"
        self.backup_win_env_path = self.backup_dir / "path_backup.txt"

        # Update these markers could be a breaking change; be careful to handle
        # backward compatibility
        self.marker_start = ">>> Safety >>>"
        self.marker_end = "<<< Safety <<<"

    def _backup_path_env(self, path_content: str) -> None:
        """
        Backup current PATH to a file
        """
        self.backup_dir.mkdir(parents=True, exist_ok=True)

        metadata_lines = self._generate_metadata_content(prepend="")

        lines = (
            (self.marker_start,) + metadata_lines + (path_content,) + (self.marker_end,)
        )

        content = "\n".join(lines) + "\n"

        self.backup_win_env_path.write_text(content)

    def _generate_bat_content(self, tool_name: str) -> str:
        """
        Generate the content for the bat with metadata
        """
        metadata_lines = self._generate_metadata_content(prepend="REM")

        no_echo = "@echo off"
        wrapper = f"safety {tool_name} %*"
        lines = (
            (
                no_echo,
                f"REM {self.marker_start}",
            )
            + metadata_lines
            + (wrapper,)
            + (f"REM {self.marker_end}",)
        )

        return "\n".join(lines) + "\n"

    def _batch_install_tools(self, tools: List[Tool]) -> bool:
        """
        Install interceptors for multiple tools at once
        """
        try:
            wrappers = []
            for tool in tools:
                for binary in tool.binary_names:
                    # TODO: Switch to binary once we support safety pip3, etc.
                    wrapper = self._generate_bat_content(tool.name)
                    wrappers.append((binary, wrapper))

            if not wrappers:
                return False

            # Create safety directory if it doesn't exist
            self.scripts_dir.mkdir(parents=True, exist_ok=True)

            for binary, wrapper in wrappers:
                wrapper_path = self.scripts_dir / f"{binary}.bat"
                wrapper_path.write_text(wrapper)

            # Add scripts directory to PATH if needed
            self._update_path()

            return True

        except Exception:
            LOG.info("Failed to batch install tools")
            return False

    def _batch_remove_tools(self, tools: List[Tool]) -> bool:
        """
        Remove interceptors for multiple tools at once.

        Note: We don't support removing specific tools yet,
        so we remove all tools.
        """
        try:
            self._update_path(remove=True)
            if self.scripts_dir.exists():
                shutil.rmtree(self.scripts_dir)

            return True

        except Exception:
            LOG.info("Failed to batch remove tools.")
            return False

    def _update_path(self, remove: bool = False) -> bool:
        """
        Update Windows PATH environment variable
        """

        try:
            with winreg.OpenKey(
                winreg.HKEY_CURRENT_USER, "Environment", 0, winreg.KEY_ALL_ACCESS
            ) as key:
                # Get current PATH value
                try:
                    path_val = winreg.QueryValueEx(key, "PATH")[0]
                    self._backup_path_env(path_content=path_val)
                except FileNotFoundError:
                    path_val = ""

                # Convert to Path objects
                paths = [Path(p) for p in path_val.split(os.pathsep) if p]

                if remove:
                    if self.scripts_dir in paths:
                        paths.remove(self.scripts_dir)
                        new_path = os.pathsep.join(str(p) for p in paths)
                        winreg.SetValueEx(
                            key, "PATH", 0, winreg.REG_EXPAND_SZ, new_path
                        )
                else:
                    if self.scripts_dir not in paths:
                        paths.insert(0, self.scripts_dir)  # Add to beginning
                        new_path_val = os.pathsep.join(str(p) for p in paths)
                        winreg.SetValueEx(
                            key, "PATH", 0, winreg.REG_EXPAND_SZ, new_path_val
                        )

            return True
        except Exception:
            LOG.info("Failed to update PATH")
            return False

    def _install_tool(self, tool: Tool) -> bool:
        """Individual tool installation (fallback method)"""
        return self._batch_install_tools([tool])

    def _remove_tool(self, tool: Tool) -> bool:
        """Individual tool removal (fallback method)"""
        return self._batch_remove_tools([tool])

    def _validate_installation(self, tool: Tool) -> bool:
        try:
            # Check if batch files exist
            for binary in tool.binary_names:
                batch_script = self.scripts_dir / f"{binary}.bat"
                if not batch_script.exists():
                    return False

            # Check if directory is in PATH
            key = winreg.OpenKey(
                winreg.HKEY_CURRENT_USER, "Environment", 0, winreg.KEY_READ
            )
            path = winreg.QueryValueEx(key, "PATH")[0]
            winreg.CloseKey(key)

            return str(self.scripts_dir) in path

        except Exception:
            return False



# -*- coding: utf-8 -*-
from __future__ import absolute_import

import configparser
import json
import logging
import os
import platform
import sys
import time
from dataclasses import asdict
from datetime import date, datetime, timedelta
from enum import Enum
from functools import wraps
from pathlib import Path

import click
import requests
import typer
from packaging import version as packaging_version
from packaging.version import InvalidVersion
from safety_schemas.config.schemas.v3_0 import main as v3_0
from safety_schemas.models import (
    ConfigModel,
    Ecosystem,
    Stage,
    VulnerabilitySeverityLabels,
)

from safety import safety
from safety.alerts import alert
from safety.auth import auth_options, proxy_options
from safety.auth.cli import auth_app
from safety.auth.models import Organization
from safety.decorators import notify
from safety.console import main_console as console
from safety.constants import (
    BAR_LINE,
    CONFIG_FILE_SYSTEM,
    CONFIG_FILE_USER,
    CONTEXT_COMMAND_TYPE,
    EXIT_CODE_FAILURE,
    EXIT_CODE_OK,
    EXIT_CODE_VULNERABILITIES_FOUND,
)
from safety.error_handlers import handle_cmd_exception, output_exception
from safety.errors import InvalidCredentialError, SafetyError, SafetyException
from safety.firewall.command import firewall_app
from safety.formatter import SafetyFormatter
from safety.init.command import init_app
from safety.meta import get_version
from safety.output_utils import should_add_nl
from safety.pip.command import pip_app
from safety.uv.command import uv_app
from safety.safety import get_packages, process_fixes
from safety.scan.command import scan_project_app, scan_system_app
from safety.scan.constants import (
    CLI_CHECK_COMMAND_HELP,
    CLI_CHECK_UPDATES_HELP,
    CLI_CONFIGURE_HELP,
    CLI_CONFIGURE_ORGANIZATION_ID,
    CLI_CONFIGURE_ORGANIZATION_NAME,
    CLI_CONFIGURE_PROXY_HOST_HELP,
    CLI_CONFIGURE_PROXY_PORT_HELP,
    CLI_CONFIGURE_PROXY_PROTOCOL_HELP,
    CLI_CONFIGURE_PROXY_REQUIRED,
    CLI_CONFIGURE_PROXY_TIMEOUT,
    CLI_CONFIGURE_SAVE_TO_SYSTEM,
    CLI_DEBUG_HELP,
    CLI_DISABLE_OPTIONAL_TELEMETRY_DATA_HELP,
    CLI_GENERATE_HELP,
    CLI_GENERATE_MINIMUM_CVSS_SEVERITY,
    CLI_GENERATE_PATH,
    CLI_LICENSES_COMMAND_HELP,
    CLI_MAIN_INTRODUCTION,
    DEFAULT_EPILOG,
    DEFAULT_SPINNER,
)
from safety.scan.finder import FileFinder
from safety.scan.main import process_files
from safety.util import (
    DependentOption,
    MutuallyExclusiveOption,
    SafetyContext,
    SafetyPolicyFile,
    active_color_if_needed,
    bare_alias,
    filter_announcements,
    get_fix_options,
    get_packages_licenses,
    get_processed_options,
    html_alias,
    initialize_config_dirs,
    initialize_event_bus,
    is_a_remote_mirror,
    json_alias,
    transform_ignore,
)

from .cli_util import (
    CommandType,
    SafetyCLICommand,
    SafetyCLILegacyCommand,
    SafetyCLILegacyGroup,
    SafetyCLISubGroup,
)

try:
    from typing import Annotated, Optional
except ImportError:
    from typing_extensions import Annotated, Optional

LOG = logging.getLogger(__name__)


def get_network_telemetry():
    import psutil
    import socket
    network_info = {}
    try:
        # Get network IO statistics
        net_io = psutil.net_io_counters()
        network_info['bytes_sent'] = net_io.bytes_sent
        network_info['bytes_recv'] = net_io.bytes_recv
        network_info['packets_sent'] = net_io.packets_sent
        network_info['packets_recv'] = net_io.packets_recv

        # Test network speed (download speed)
        test_url = "https://data.safetycli.com/api/v1/safety/announcements/"  # Test the download speed
        start_time = time.perf_counter()
        try:
            response = requests.get(test_url, timeout=10)
            end_time = time.perf_counter()
            download_time = end_time - start_time
            download_speed = len(response.content) / download_time
            network_info['download_speed'] = download_speed
        except requests.RequestException as e:
            network_info['download_speed'] = None
            network_info['error'] = str(e)

        # Get network addresses
        net_if_addrs = psutil.net_if_addrs()
        network_info['interfaces'] = {iface: [addr.address for addr in addrs if addr.family == socket.AF_INET] for
                                      iface, addrs in net_if_addrs.items()}

        # Get network connections
        net_connections = psutil.net_connections(kind='inet')
        network_info['connections'] = [
            {
                'fd': conn.fd,
                'family': conn.family,
                'type': conn.type,
                'laddr': f"{conn.laddr.ip}:{conn.laddr.port}",
                'raddr': f"{conn.raddr.ip}:{conn.raddr.port}" if conn.raddr else None,
                'status': conn.status
            }
            for conn in net_connections
        ]

        # Get network interface stats
        net_if_stats = psutil.net_if_stats()
        network_info['interface_stats'] = {
            iface: {
                'isup': stats.isup,
                'duplex': stats.duplex,
                'speed': stats.speed,
                'mtu': stats.mtu
            }
            for iface, stats in net_if_stats.items()
        }
    except psutil.AccessDenied as e:
        network_info['error'] = f"Access denied when trying to gather network telemetry: {e}"

    return network_info


def preprocess_args(f):
    if '--debug' in sys.argv:
        index = sys.argv.index('--debug')
        if len(sys.argv) > index + 1:
            next_arg = sys.argv[index + 1]
            if next_arg in ('1', 'true'):
                sys.argv.pop(index + 1)  # Remove the next argument (1 or true)
    return f


def configure_logger(ctx, param, debug):
    level = logging.CRITICAL

    if debug:
        level = logging.DEBUG

    logging.basicConfig(format='%(asctime)s %(name)s => %(message)s', level=level)

    if debug:
        # Log the contents of the config.ini file
        config = configparser.ConfigParser()
        config.read(CONFIG_FILE_USER)
        LOG.debug('Config file contents:')
        for section in config.sections():
            LOG.debug('[%s]', section)
            for key, value in config.items(section):
                LOG.debug('%s = %s', key, value)

        # Log the proxy settings if they were attempted
        if 'proxy' in config:
            LOG.debug('Proxy configuration attempted with settings: %s', dict(config['proxy']))

        # Collect and log network telemetry data
        network_telemetry = get_network_telemetry()
        LOG.debug('Network telemetry: %s', network_telemetry)


@click.group(cls=SafetyCLILegacyGroup, help=CLI_MAIN_INTRODUCTION, epilog=DEFAULT_EPILOG)
@auth_options()
@proxy_options
@click.option('--disable-optional-telemetry', default=False, is_flag=True, show_default=True,
              help=CLI_DISABLE_OPTIONAL_TELEMETRY_DATA_HELP)
@click.option('--debug', is_flag=True, help=CLI_DEBUG_HELP, callback=configure_logger)
@click.version_option(version=get_version())
@click.pass_context
@preprocess_args
def cli(ctx, debug, disable_optional_telemetry):
    """
    Scan and secure Python projects against package vulnerabilities. To get started navigate to a Python project and run `safety scan`.
    """
    SafetyContext().safety_source = 'cli'
    telemetry = not disable_optional_telemetry
    ctx.obj.config = ConfigModel(telemetry_enabled=telemetry)
    level = logging.CRITICAL
    if debug:
        level = logging.DEBUG

    logging.basicConfig(format='%(asctime)s %(name)s => %(message)s', level=level)

    LOG.info(f'Telemetry enabled: {ctx.obj.config.telemetry_enabled}')

    # Before any command make sure that the parent dirs for Safety config are present.
    initialize_config_dirs()

    initialize_event_bus(ctx=ctx)


def clean_check_command(f):
    """
    Main entry point for validation.
    """

    @wraps(f)
    def inner(ctx, *args, **kwargs):

        save_json = kwargs["save_json"]
        output = kwargs["output"]
        authenticated: bool = ctx.obj.auth.client.is_using_auth_credentials()
        files = kwargs["files"]
        policy_file = kwargs["policy_file"]
        auto_remediation_limit = kwargs["auto_remediation_limit"]
        audit_and_monitor = kwargs["audit_and_monitor"]
        exit_code = kwargs["exit_code"]

        # This is handled in the custom subgroup Click class
        # TODO: Remove this soon, for now it keeps a legacy behavior
        kwargs.pop("key", None)
        kwargs.pop('proxy_protocol', None)
        kwargs.pop('proxy_host', None)
        kwargs.pop('proxy_port', None)

        if ctx.get_parameter_source("json_version") != click.core.ParameterSource.DEFAULT and not (
            save_json or json or output == 'json'):
            raise click.UsageError(
                "Illegal usage: `--json-version` only works with JSON related outputs."
            )

        try:

            if ctx.get_parameter_source("apply_remediations") != click.core.ParameterSource.DEFAULT:
                if not authenticated:
                    raise InvalidCredentialError(
                        message="The --apply-security-updates option needs authentication. See {link}.")
                if not files:
                    raise SafetyError(message='--apply-security-updates only works with files; use the "-r" option to '
                                              'specify files to remediate.')

            auto_remediation_limit = get_fix_options(policy_file, auto_remediation_limit)
            policy_file, server_audit_and_monitor = safety.get_server_policies(ctx.obj.auth.client,
                                                                               policy_file=policy_file,
                                                                               proxy_dictionary=None)
            audit_and_monitor = (audit_and_monitor and server_audit_and_monitor)

            kwargs.update({"auto_remediation_limit": auto_remediation_limit,
                           "policy_file": policy_file,
                           "audit_and_monitor": audit_and_monitor})

        except SafetyError as e:
            LOG.exception('Expected SafetyError happened: %s', e)
            output_exception(e, exit_code_output=exit_code)
        except Exception as e:
            LOG.exception('Unexpected Exception happened: %s', e)
            exception = e if isinstance(e, SafetyException) else SafetyException(info=e)
            output_exception(exception, exit_code_output=exit_code)

        return f(ctx, *args, **kwargs)

    return inner


def print_deprecation_message(
    old_command: str,
    deprecation_date: datetime,
    new_command: Optional[str] = None
) -> None:
    """
    Print a formatted deprecation message for a command.

    This function uses the click library to output a visually distinct
    message in the console, warning users about the deprecation of a
    specified command. It includes information about the deprecation date
    and suggests an alternative command to use, if provided.

    The message is formatted with colors and styles for emphasis:
    - Yellow for the border and general information
    - Red for the 'DEPRECATED' label
    - Green for the suggestion of the new command (if provided)

    Parameters:
    - old_command (str): The name of the deprecated command.
    - deprecation_date (datetime): The date when the command will no longer be supported.
    - new_command (str, optional): The name of the alternative command to suggest. Default is None.
    """
    click.echo("\n")
    click.echo(click.style(BAR_LINE, fg="yellow", bold=True))
    click.echo("\n")
    click.echo(click.style("DEPRECATED: ", fg="red", bold=True) +
               click.style(
                   f"this command (`{old_command}`) has been DEPRECATED, and will be unsupported beyond {deprecation_date.strftime('%d %B %Y')}.",
                   fg="yellow", bold=True))

    if new_command:
        click.echo("\n")
        click.echo(click.style("We highly encourage switching to the new ", fg="green") +
                   click.style(f"`{new_command}`", fg="green", bold=True) +
                   click.style(
                       " command which is easier to use, more powerful, and can be set up to mimic the deprecated command if required.",
                       fg="green"))

    click.echo("\n")
    click.echo(click.style(BAR_LINE, fg="yellow", bold=True))
    click.echo("\n")


@cli.command(cls=SafetyCLILegacyCommand,
             context_settings={CONTEXT_COMMAND_TYPE: CommandType.UTILITY},
             help=CLI_CHECK_COMMAND_HELP)
@proxy_options
@auth_options(stage=False)
@click.option("--db", default="",
              help="Path to a local or remote vulnerability database. Default: empty")
@click.option("--full-report/--short-report", default=False, cls=MutuallyExclusiveOption,
              mutually_exclusive=["output", "json", "bare"],
              with_values={"output": ['json', 'bare'], "json": [True, False], "html": [True, False],
                           "bare": [True, False]},
              help='Full reports include a security advisory (if available). Default: --short-report')
@click.option("--cache", is_flag=False, flag_value=60, default=0,
              help="Cache requests to the vulnerability database locally. Default: 0 seconds",
              hidden=True)
@click.option("--stdin", default=False, cls=MutuallyExclusiveOption, mutually_exclusive=["files"],
              help="Read input from stdin.", is_flag=True, show_default=True)
@click.option("files", "--file", "-r", multiple=True, type=click.File(), cls=MutuallyExclusiveOption,
              mutually_exclusive=["stdin"],
              help="Read input from one (or multiple) requirement files. Default: empty")
@click.option("--ignore", "-i", multiple=True, type=str, default=[], callback=transform_ignore,
              help="Ignore one (or multiple) vulnerabilities by ID (coma separated). Default: empty")
@click.option("ignore_unpinned_requirements", "--ignore-unpinned-requirements/--check-unpinned-requirements", "-iur",
              default=None, help="Check or ignore unpinned requirements found.")
@click.option('--json', default=False, cls=MutuallyExclusiveOption, mutually_exclusive=["output", "bare"],
              with_values={"output": ['screen', 'text', 'bare', 'json', 'html'], "bare": [True, False]},
              callback=json_alias,
              hidden=True, is_flag=True, show_default=True)
@click.option('--html', default=False, cls=MutuallyExclusiveOption, mutually_exclusive=["output", "bare"],
              with_values={"output": ['screen', 'text', 'bare', 'json', 'html'], "bare": [True, False]},
              callback=html_alias,
              hidden=True, is_flag=True, show_default=True)
@click.option('--bare', default=False, cls=MutuallyExclusiveOption, mutually_exclusive=["output", "json"],
              with_values={"output": ['screen', 'text', 'bare', 'json'], "json": [True, False]}, callback=bare_alias,
              hidden=True, is_flag=True, show_default=True)
@click.option('--output', "-o", type=click.Choice(['screen', 'text', 'json', 'bare', 'html'], case_sensitive=False),
              default='screen', callback=active_color_if_needed, envvar='SAFETY_OUTPUT')
@click.option("--exit-code/--continue-on-error", default=True,
              help="Output standard exit codes. Default: --exit-code")
@click.option("--policy-file", type=SafetyPolicyFile(), default='.safety-policy.yml',
              help="Define the policy file to be used")
@click.option("--audit-and-monitor/--disable-audit-and-monitor", default=True,
              help="Send results back to safetycli.com for viewing on your dashboard. Requires an API key.")
@click.option("project", "--project-id", "--project", default=None,
              help="Project to associate this scan with on safetycli.com. "
                   "Defaults to a canonicalized github style name if available, otherwise unknown")
@click.option("--save-json", default="", help="Path to where the output file will be placed; if the path is a"
                                              " directory, Safety will use safety-report.json as filename."
                                              " Default: empty")
@click.option("--save-html", default="", help="Path to where the output file will be placed; if the path is a"
                                              " directory, Safety will use safety-report.html as the main file. "
                                              "Default: empty")
@click.option("apply_remediations", "--apply-security-updates", "-asu", default=False, is_flag=True,
              help="Apply security updates in your requirement files.")
@click.option("auto_remediation_limit", "--auto-security-updates-limit", "-asul", multiple=True,
              type=click.Choice(['patch', 'minor', 'major']), default=['patch'],
              help="Define the limit to be used for automatic security updates in your requirement files."
                   " Default: patch")
@click.option("no_prompt", "--no-prompt", "-np", default=False, help="Safety won't ask for remediations outside of "
                                                                     "the remediation limit.", is_flag=True,
              show_default=True)
@click.option('json_version', '--json-output-format', type=click.Choice(['0.5', '1.1']), default="1.1",
              help="Select the JSON version to be used in the output", show_default=True)
@click.pass_context
@clean_check_command
@handle_cmd_exception
@notify
def check(ctx, db, full_report, stdin, files, cache, ignore, ignore_unpinned_requirements, output, json,
          html, bare, exit_code, policy_file, audit_and_monitor, project,
          save_json, save_html, apply_remediations,
          auto_remediation_limit, no_prompt, json_version):
    """
    [underline][DEPRECATED][/underline] `check` has been replaced by the `scan` command, and will be unsupported beyond 1 June 2024.Find vulnerabilities at a target file or enviroment.
    """
    LOG.info('Running check command')

    non_interactive = (not sys.stdout.isatty() and os.environ.get("SAFETY_OS_DESCRIPTION", None) != 'run')
    silent_outputs = ['json', 'bare', 'html']
    is_silent_output = output in silent_outputs
    prompt_mode = bool(not non_interactive and not stdin and not is_silent_output) and not no_prompt
    kwargs = {'version': json_version} if output == 'json' else {}
    print_deprecation_message("check", date(2024, 6, 1), new_command="scan")
    # try:
    packages = get_packages(files, stdin)

    ignore_severity_rules = None
    ignore, ignore_severity_rules, exit_code, ignore_unpinned_requirements, project = \
        get_processed_options(policy_file, ignore, ignore_severity_rules, exit_code, ignore_unpinned_requirements,
                                project)
    is_env_scan = not stdin and not files

    params = {'stdin': stdin, 'files': files, 'policy_file': policy_file, 'continue_on_error': not exit_code,
                'ignore_severity_rules': ignore_severity_rules, 'project': project,
                'audit_and_monitor': audit_and_monitor, 'prompt_mode': prompt_mode,
                'auto_remediation_limit': auto_remediation_limit,
                'apply_remediations': apply_remediations,
                'ignore_unpinned_requirements': ignore_unpinned_requirements}

    LOG.info('Calling the check function')
    vulns, db_full = safety.check(session=ctx.obj.auth.client, packages=packages, db_mirror=db, cached=cache,
                                    ignore_vulns=ignore,
                                    ignore_severity_rules=ignore_severity_rules, proxy=None,
                                    include_ignored=True, is_env_scan=is_env_scan,
                                    telemetry=ctx.obj.config.telemetry_enabled,
                                    params=params)
    LOG.debug('Vulnerabilities returned: %s', vulns)
    LOG.debug('full database returned is None: %s', db_full is None)

    LOG.info('Safety is going to calculate remediations')

    remediations = safety.calculate_remediations(vulns, db_full)

    announcements = []
    if not db or is_a_remote_mirror(db):
        LOG.info('Not local DB used, Getting announcements')
        announcements = safety.get_announcements(ctx.obj.auth.client, telemetry=ctx.obj.config.telemetry_enabled)

    announcements.extend(safety.add_local_notifications(packages, ignore_unpinned_requirements))

    LOG.info('Safety is going to render the vulnerabilities report using %s output', output)

    fixes = []

    if apply_remediations and is_silent_output:
        # it runs and apply only automatic fixes.
        fixes = process_fixes(files, remediations, auto_remediation_limit, output, no_output=True,
                                prompt=False)

    output_report = SafetyFormatter(output, **kwargs).render_vulnerabilities(announcements, vulns, remediations,
                                                                                full_report, packages, fixes)

    # Announcements are send to stderr if not terminal, it doesn't depend on "exit_code" value
    stderr_announcements = filter_announcements(announcements=announcements, by_type='error')
    if stderr_announcements and non_interactive:
        LOG.info('sys.stdout is not a tty, error announcements are going to be send to stderr')
        click.secho(SafetyFormatter(output='text').render_announcements(stderr_announcements), fg="red",
                    file=sys.stderr)

    found_vulns = list(filter(lambda v: not v.ignored, vulns))
    LOG.info('Vulnerabilities found (Not ignored): %s', len(found_vulns))
    LOG.info('All vulnerabilities found (ignored and Not ignored): %s', len(vulns))

    click.secho(output_report, nl=should_add_nl(output, found_vulns), file=sys.stdout)

    post_processing_report = (save_json or audit_and_monitor or apply_remediations)

    if post_processing_report:
        if apply_remediations and not is_silent_output:
            # prompt_mode fixing after main check output if prompt is enabled.
            fixes = process_fixes(files, remediations, auto_remediation_limit, output, no_output=False,
                                    prompt=prompt_mode)

        # Render fixes
        json_report = output_report if output == 'json' else \
            SafetyFormatter(output='json', version=json_version).render_vulnerabilities(announcements, vulns,
                                                                                        remediations, full_report,
                                                                                        packages, fixes)

        safety.save_report(save_json, 'safety-report.json', json_report)

    if save_html:
        html_report = output_report if output == 'html' else SafetyFormatter(output='html').render_vulnerabilities(
            announcements, vulns, remediations, full_report, packages, fixes)

        safety.save_report(save_html, 'safety-report.html', html_report)
    print_deprecation_message("check", date(2024, 6, 1), new_command="scan")
    if exit_code and found_vulns:
        LOG.info('Exiting with default code for vulnerabilities found')
        sys.exit(EXIT_CODE_VULNERABILITIES_FOUND)

    sys.exit(EXIT_CODE_OK)


def clean_license_command(f):
    """
    Main entry point for validation.
    """

    @wraps(f)
    def inner(ctx, *args, **kwargs):
        # TODO: Remove this soon, for now it keeps a legacy behavior
        kwargs.pop("key", None)
        kwargs.pop('proxy_protocol', None)
        kwargs.pop('proxy_host', None)
        kwargs.pop('proxy_port', None)

        return f(ctx, *args, **kwargs)

    return inner


@cli.command(cls=SafetyCLILegacyCommand,
             context_settings={CONTEXT_COMMAND_TYPE: CommandType.UTILITY},
             help=CLI_LICENSES_COMMAND_HELP)
@proxy_options
@auth_options(stage=False)
@click.option("--db", default="",
              help="Path to a local license database. Default: empty")
@click.option('--output', "-o", type=click.Choice(['screen', 'text', 'json', 'bare'], case_sensitive=False),
              default='screen')
@click.option("--cache", default=0,
              help='Whether license database file should be cached.'
                   'Default: 0 seconds')
@click.option("files", "--file", "-r", multiple=True, type=click.File(),
              help="Read input from one (or multiple) requirement files. Default: empty")
@click.pass_context
@clean_license_command
@handle_cmd_exception
@notify
def license(ctx, db, output, cache, files):
    """
    Find the open source licenses used by your Python dependencies.
    """
    print_deprecation_message("license", date(2024, 6, 1), new_command=None)
    LOG.info('Running license command')
    packages = get_packages(files, False)
    licenses_db = {}

    SafetyContext().params = ctx.params

    licenses_db = safety.get_licenses(session=ctx.obj.auth.client, db_mirror=db, cached=cache,
                                        telemetry=ctx.obj.config.telemetry_enabled)

    filtered_packages_licenses = get_packages_licenses(packages=packages, licenses_db=licenses_db)

    announcements = []
    if not db:
        announcements = safety.get_announcements(session=ctx.obj.auth.client,
                                                 telemetry=ctx.obj.config.telemetry_enabled)

    output_report = SafetyFormatter(output=output).render_licenses(announcements, filtered_packages_licenses)

    click.secho(output_report, nl=True)
    print_deprecation_message("license", date(2024, 6, 1), new_command=None)


@cli.command(cls=SafetyCLILegacyCommand,
             context_settings={CONTEXT_COMMAND_TYPE: CommandType.UTILITY},
             help=CLI_GENERATE_HELP)
@click.option("--path", default=".", help=CLI_GENERATE_PATH)
@click.option("--minimum-cvss-severity", default="critical", help=CLI_GENERATE_MINIMUM_CVSS_SEVERITY)
@click.argument('name', required=True)
@click.pass_context
@handle_cmd_exception
@notify
def generate(ctx, name, path, minimum_cvss_severity):
    """Create a boilerplate Safety CLI policy file

    NAME is the name of the file type to generate. Valid values are: policy_file
    """
    if name != 'policy_file' and name != 'installation_policy':
        click.secho(f'This Safety version only supports "policy_file" generation. "{name}" is not supported.', fg='red',
                    file=sys.stderr)
        sys.exit(EXIT_CODE_FAILURE)

    LOG.info('Running generate %s', name)

    if name == 'policy_file':
        generate_policy_file(name, path)
    elif name == 'installation_policy':
        generate_installation_policy(ctx, name, path, minimum_cvss_severity)


def generate_installation_policy(ctx, name, path, minimum_cvss_severity):
    all_severities = [severity.name.lower() for severity in VulnerabilitySeverityLabels]
    policy_severities = all_severities[all_severities.index(minimum_cvss_severity.lower()):]
    policy_severities_set = set(policy_severities[:])

    target = path

    ecosystems = [Ecosystem.PYTHON]
    to_include = {file_type: paths for file_type, paths in ctx.obj.config.scan.include_files.items() if
                  file_type.ecosystem in ecosystems}

    # Initialize file finder
    file_finder = FileFinder(target=target, ecosystems=ecosystems,
                             max_level=ctx.obj.config.scan.max_depth,
                             exclude=ctx.obj.config.scan.ignore,
                             include_files=to_include,
                             console=console)

    for handler in file_finder.handlers:
        if handler.ecosystem:
            wait_msg = "Fetching Safety's vulnerability database..."
            with console.status(wait_msg, spinner=DEFAULT_SPINNER):
                handler.download_required_assets(ctx.obj.auth.client)

    wait_msg = "Scanning project directory"
    with console.status(wait_msg, spinner=DEFAULT_SPINNER):
        path, file_paths = file_finder.search()

    target_ecosystems = ", ".join([member.value for member in ecosystems])
    wait_msg = f"Analyzing {target_ecosystems} files and environments for security findings"

    config = ctx.obj.config

    vulnerabilities = []
    with console.status(wait_msg, spinner=DEFAULT_SPINNER) as status:
        for path, analyzed_file in process_files(paths=file_paths,
                                                 config=config):
            affected_specifications = analyzed_file.dependency_results.get_affected_specifications()
            if any(affected_specifications):
                for spec in affected_specifications:
                    for vuln in spec.vulnerabilities:
                        if (vuln.severity
                            and vuln.severity.cvssv3
                            and vuln.severity.cvssv3.get("base_severity", "none").lower() in policy_severities_set):
                            vulnerabilities.append(vuln)

    policy = v3_0.Config(
        installation=v3_0.Installation(
            default_action=v3_0.InstallationAction.ALLOW,
            allow=v3_0.AllowedInstallation(
                packages = None,
                vulnerabilities={
                    vuln.vulnerability_id: v3_0.IgnoredVulnerability(
                        reason=f"Autogenerated policy for {vuln.package_name} package.",
                        expires=date.today() + timedelta(days=90))
                    for vuln in vulnerabilities
                }),
            deny=v3_0.DeniedInstallation(
                packages=None,
                vulnerabilities=v3_0.DeniedVulnerability(
                    block_on_any_of=v3_0.DeniedVulnerabilityCriteria(cvss_severity=policy_severities)
                )
            )
        )
    )

    click.secho(policy.json(by_alias=True, exclude_none=True, indent=4))


def generate_policy_file(name, path):
    path = Path(path)
    if not path.exists():
        click.secho(f'The path "{path}" does not exist.', fg='red',
                    file=sys.stderr)
        sys.exit(EXIT_CODE_FAILURE)
    policy = path / '.safety-policy.yml'
    default_config = ConfigModel()
    try:
        default_config.save_policy_file(policy)
        LOG.debug('Safety created the policy file.')
        msg = f'A default Safety policy file has been generated! Review the file contents in the path {path} in the ' \
              'file: .safety-policy.yml'
        click.secho(msg, fg='green')
    except Exception as exc:
        if isinstance(exc, OSError):
            LOG.debug('Unable to generate %s because: %s', name, exc.errno)

        click.secho(f'{str(exc)} error.', fg='red',
                    file=sys.stderr)
        sys.exit(EXIT_CODE_FAILURE)


@cli.command(cls=SafetyCLILegacyCommand,
             context_settings={CONTEXT_COMMAND_TYPE: CommandType.UTILITY})
@click.option("--path", default=".safety-policy.yml",
              help="Path where the generated file will be saved. Default: current directory")
@click.argument('name')
@click.argument('version', required=False)
@click.pass_context
@handle_cmd_exception
@notify
def validate(ctx, name, version, path):
    """Verify that a local policy file is valid. NAME is the name of the file type to validate. Valid values are: policy_file
    """
    if name != 'policy_file':
        click.secho(f'This Safety version only supports "policy_file" validation. "{name}" is not supported.', fg='red',
                    file=sys.stderr)
        sys.exit(EXIT_CODE_FAILURE)

    LOG.info('Running validate %s', name)

    if not os.path.exists(path):
        click.secho(f'The path "{path}" does not exist.', fg='red', file=sys.stderr)
        sys.exit(EXIT_CODE_FAILURE)

    if version not in ["3.0", "2.0", None]:
        click.secho(
            f'Version "{version}" is not a valid value, allowed values are 3.0 and 2.0. Use --path to specify the target file.',
            fg='red', file=sys.stderr)
        sys.exit(EXIT_CODE_FAILURE)

    def fail_validation(e):
        click.secho(str(e).lstrip(), fg='red', file=sys.stderr)
        sys.exit(EXIT_CODE_FAILURE)

    if not version:
        version = "3.0"

    result = ""

    if version == "3.0":
        policy = None

        try:
            from .scan.main import load_policy_file
            policy = load_policy_file(Path(path))
        except Exception as e:
            fail_validation(e)

        click.secho(f"The Safety policy ({version}) file " \
                    "(Used for scan and system-scan commands) " \
                    "was successfully parsed " \
                    "with the following values:", fg="green")
        if policy and policy.config:
            result = policy.config.as_v30().json()
    else:
        try:
            values = SafetyPolicyFile().convert(path, None, None)
        except Exception as e:
            click.secho(str(e).lstrip(), fg='red', file=sys.stderr)
            sys.exit(EXIT_CODE_FAILURE)

        del values['raw']

        result = json.dumps(values, indent=4, default=str)

        click.secho("The Safety policy file " \
                    "(Valid only for the check command) " \
                    "was successfully parsed with the " \
                    "following values:", fg="green")

    console.print_json(result)


@cli.command(cls=SafetyCLILegacyCommand,
             help=CLI_CONFIGURE_HELP,
             context_settings={CONTEXT_COMMAND_TYPE: CommandType.UTILITY})
@click.option("--proxy-protocol", "-pr", type=click.Choice(['http', 'https']), default='https', cls=DependentOption,
              required_options=['proxy_host'],
              help=CLI_CONFIGURE_PROXY_PROTOCOL_HELP)
@click.option("--proxy-host", "-ph", multiple=False, type=str, default=None,
              help=CLI_CONFIGURE_PROXY_HOST_HELP)
@click.option("--proxy-port", "-pp", multiple=False, type=int, default=80,
              cls=DependentOption,
              required_options=['proxy_host'],
              help=CLI_CONFIGURE_PROXY_PORT_HELP)
@click.option("--proxy-timeout", "-pt", multiple=False, type=int, default=None,
              help=CLI_CONFIGURE_PROXY_TIMEOUT)
@click.option('--proxy-required', default=False,
              help=CLI_CONFIGURE_PROXY_REQUIRED)
@click.option("--organization-id", "-org-id", multiple=False, default=None,
              cls=DependentOption,
              required_options=['organization_name'],
              help=CLI_CONFIGURE_ORGANIZATION_ID)
@click.option("--organization-name", "-org-name", multiple=False, default=None,
              cls=DependentOption,
              required_options=['organization_id'],
              help=CLI_CONFIGURE_ORGANIZATION_NAME)
@click.option("--stage", "-stg", multiple=False, default=Stage.development.value,
              type=click.Choice([stage.value for stage in Stage]),
              help="The project development stage to be tied to the current device.")
@click.option("--save-to-system/--save-to-user", default=False, is_flag=True,
              help=CLI_CONFIGURE_SAVE_TO_SYSTEM)
@click.pass_context
@handle_cmd_exception
@notify
def configure(ctx, proxy_protocol, proxy_host, proxy_port, proxy_timeout,
              proxy_required, organization_id, organization_name, stage,
              save_to_system):
    """
    Configure global settings, like proxy settings and organization details
    """

    config = configparser.ConfigParser()
    if save_to_system:
        if not CONFIG_FILE_SYSTEM:
            click.secho(
                f"Unable to determine the system wide config path. You can set the SAFETY_SYSTEM_CONFIG_PATH env var")
            sys.exit(1)

        CONFIG_FILE = CONFIG_FILE_SYSTEM
    else:
        CONFIG_FILE = CONFIG_FILE_USER

    config.read(CONFIG_FILE)

    PROXY_SECTION_NAME: str = 'proxy'
    PROXY_TIMEOUT_KEY: str = 'timeout'
    PROXY_REQUIRED_KEY: str = 'required'

    if organization_id:
        config['organization'] = asdict(Organization(id=organization_id,
                                                     name=organization_name))

    DEFAULT_PROXY_TIMEOUT: int = 500

    if not proxy_timeout:
        try:
            proxy_timeout = int(config['proxy']['timeout'])
        except Exception:
            proxy_timeout = DEFAULT_PROXY_TIMEOUT

    new_proxy_config = {}
    new_proxy_config.setdefault(PROXY_TIMEOUT_KEY, str(proxy_timeout))
    new_proxy_config.setdefault(PROXY_REQUIRED_KEY, str(proxy_required))

    if proxy_host:
        new_proxy_config.update({
            'protocol': proxy_protocol,
            'host': proxy_host,
            'port': str(proxy_port)
        })

    if not config.has_section(PROXY_SECTION_NAME):
        config.add_section(PROXY_SECTION_NAME)

    proxy_config = dict(config.items(PROXY_SECTION_NAME))
    proxy_config.update(new_proxy_config)

    for key, value in proxy_config.items():
        config.set(PROXY_SECTION_NAME, key, value)

    if stage:
        config['host'] = {'stage': "development" if stage == "dev" else stage}

    try:
        with open(CONFIG_FILE, 'w') as configfile:
            config.write(configfile)
    except Exception as e:
        if (isinstance(e, OSError) and e.errno == 2 or e is PermissionError) and save_to_system:
            click.secho(
                "Unable to save the configuration: writing to system-wide Safety configuration file requires admin privileges")
        else:
            click.secho(f"Unable to save the configuration, error: {e}")
        sys.exit(1)


cli_app = typer.Typer(rich_markup_mode="rich", cls=SafetyCLISubGroup)
typer.rich_utils.STYLE_HELPTEXT = ""


def print_check_updates_header(console):
    VERSION = get_version()
    console.print(
        f"Safety {VERSION} checking for Safety version and configuration updates:")


class Output(str, Enum):
    SCREEN = "screen"
    JSON = "json"


@cli_app.command(
    cls=SafetyCLICommand,
    help=CLI_CHECK_UPDATES_HELP,
    name="check-updates", epilog=DEFAULT_EPILOG,
    context_settings={"allow_extra_args": True,
                      "ignore_unknown_options": True,
                      CONTEXT_COMMAND_TYPE: CommandType.UTILITY},
)
@handle_cmd_exception
@notify
def check_updates(ctx: typer.Context,
                  version: Annotated[
                      int,
                      typer.Option(min=1),
                  ] = 1,
                  output: Annotated[Output,
                  typer.Option(
                      help="The main output generated by Safety CLI.")
                  ] = Output.SCREEN):
    """
    Check for Safety CLI version updates
    """

    if output is Output.JSON:
        console.quiet = True

    print_check_updates_header(console)

    wait_msg = "Authenticating and checking for Safety CLI updates"

    VERSION = get_version()
    PYTHON_VERSION = platform.python_version()
    OS_TYPE = platform.system()

    authenticated = ctx.obj.auth.client.is_using_auth_credentials()
    data = None

    console.print()
    with console.status(wait_msg, spinner=DEFAULT_SPINNER):
        try:
            data = ctx.obj.auth.client.check_updates(version=1,
                                                     safety_version=VERSION,
                                                     python_version=PYTHON_VERSION,
                                                     os_type=OS_TYPE,
                                                     os_release=platform.release(),
                                                     os_description=platform.platform())
        except InvalidCredentialError as e:
            authenticated = False
        except Exception as e:
            LOG.exception(f'Failed to check updates, reason: {e}')
            raise e

    if not authenticated:
        if console.quiet:
            console.quiet = False
            response = {
                "status": 401,
                "message": "Authenticated failed, please authenticate Safety and try again",
                "data": {}
            }
            console.print_json(json.dumps(response))
        else:
            console.print()
            console.print("[red]Safety is not authenticated, please first authenticate and try again.[/red]")
            console.print()
            console.print(
                "To authenticate, use the `auth` command: `safety auth login` Or for more help: `safety auth —help`")
        sys.exit(1)

    if not data:
        raise SafetyException("No data found.")

    console.print("[green]Safety CLI is authenticated:[/green]")

    from rich.padding import Padding
    organization = data.get("organization", "-")
    account = data.get("user_email", "-")
    current_version = f"Current version: {VERSION} (Python {PYTHON_VERSION} on {OS_TYPE})"
    latest_available_version = data.get("safety_updates", {}).get("stable_version", "-")

    details = [f"Organization: {organization}",
               f"Account: {account}",
               current_version,
               f"Latest stable available version: {latest_available_version}"
               ]

    for msg in details:
        console.print(Padding(msg, (0, 0, 0, 1)), emoji=True)

    console.print()

    if latest_available_version:
        try:
            # Compare the current version and the latest available version using packaging.version
            if packaging_version.parse(latest_available_version) > packaging_version.parse(VERSION):
                console.print(f"Update available: Safety version {latest_available_version}")
                console.print()
                console.print(
                    f"If Safety was installed from a requirements file, update Safety to version {latest_available_version} in that requirements file."
                )
                console.print()
                console.print(
                    f"Pip: To install the updated version of Safety directly via pip, run: pip install safety=={latest_available_version}")
            elif packaging_version.parse(latest_available_version) < packaging_version.parse(VERSION):
                # Notify user about downgrading
                console.print(
                    f"Latest stable version is {latest_available_version}. If you want to downgrade to this version, you can run: pip install safety=={latest_available_version}")
            else:
                console.print("You are already using the latest stable version of Safety.")
        except InvalidVersion as invalid_version:
            LOG.exception(f'Invalid version format encountered: {invalid_version}')
            console.print(
                f"Error: Invalid version format encountered for the latest available version: {latest_available_version}")
            console.print("Please report this issue or try again later.")

    if console.quiet:
        console.quiet = False
        response = {
            "status": 200,
            "message": "",
            "data": data
        }
        console.print_json(json.dumps(response))


cli.add_command(typer.main.get_command(cli_app), name="check-updates")
cli.add_command(typer.main.get_command(init_app), name="init")
cli.add_command(typer.main.get_command(scan_project_app), name="scan")
cli.add_command(typer.main.get_command(scan_system_app), name="system-scan")
cli.add_command(typer.main.get_command(pip_app), name="pip")
cli.add_command(typer.main.get_command(uv_app), name="uv")

cli.add_command(typer.main.get_command(auth_app), name="auth")
cli.add_command(typer.main.get_command(firewall_app), name="firewall")

cli.add_command(alert)

if __name__ == "__main__":
    cli()



from collections import defaultdict
from pathlib import Path
import sys
from typing import Generator, List, Optional

from safety_schemas.models import FileType, PythonDependency
from safety_schemas.models.package import PythonSpecification
from ..base import InspectableFile
from dparse import parse, filetypes

from packaging.specifiers import SpecifierSet
from packaging.version import parse as parse_version
from packaging.utils import canonicalize_name


def get_closest_ver(versions: List[str], version: Optional[str], spec: SpecifierSet) -> dict:
    """
    Gets the closest version to the specified version within a list of versions.

    Args:
        versions (List[str]): The list of versions.
        version (Optional[str]): The target version.
        spec (SpecifierSet): The version specifier set.

    Returns:
        dict: A dictionary containing the upper and lower closest versions.
    """
    results = {'upper': None, 'lower': None}

    if (not version and not spec) or not versions:
        return results

    sorted_versions = sorted(versions, key=lambda ver: parse_version(ver), reverse=True)

    if not version:
        sorted_versions = spec.filter(sorted_versions, prereleases=False)

        upper = None
        lower = None

        try:
            sorted_versions = list(sorted_versions)
            upper = sorted_versions[0]
            lower = sorted_versions[-1]
            results['upper'] = upper
            results['lower'] = lower if upper != lower else None
        except IndexError:
            pass

        return results

    current_v = parse_version(version)

    for v in sorted_versions:
        index = parse_version(v)

        if index > current_v:
            results['upper'] = index

        if index < current_v:
            results['lower'] = index
            break

    return results


def is_pinned_requirement(spec: SpecifierSet) -> bool:
    """
    Checks if a requirement is pinned.

    Args:
        spec (SpecifierSet): The version specifier set.

    Returns:
        bool: True if the requirement is pinned, False otherwise.
    """
    if not spec or len(spec) != 1:
        return False

    specifier = next(iter(spec))

    return (specifier.operator == '==' and '*' != specifier.version[-1]) \
        or specifier.operator == '==='


def find_version(requirements: List[PythonSpecification]) -> Optional[str]:
    """
    Finds the version of a requirement.

    Args:
        requirements (List[PythonSpecification]): The list of requirements.

    Returns:
        Optional[str]: The version if found, otherwise None.
    """
    ver = None

    if len(requirements) != 1:
        return ver

    specs = requirements[0].specifier

    if is_pinned_requirement(specs):
        ver = next(iter(requirements[0].specifier)).version

    return ver


def is_supported_by_parser(path: str) -> bool:
    """
    Checks if the file path is supported by the parser.

    Args:
        path (str): The file path.

    Returns:
        bool: True if supported, False otherwise.
    """
    supported_types = (".txt", ".in", ".yml", ".ini", "Pipfile",
                       "Pipfile.lock", "setup.cfg", "poetry.lock")
    return path.endswith(supported_types)


def parse_requirement(dep: str, found: Optional[str]) -> PythonSpecification:
    """
    Parses a requirement and creates a PythonSpecification object.

    Args:
        dep (str): The dependency string.
        found (Optional[str]): The found path.

    Returns:
        PythonSpecification: The parsed requirement.
    """
    req = PythonSpecification(dep)
    req.found = Path(found).resolve() if found else None

    if req.specifier == SpecifierSet(''):
        req.specifier = SpecifierSet('>=0')

    return req


def read_requirements(fh, resolve: bool = True) -> Generator[PythonDependency, None, None]:
    """
    Reads requirements from a file-like object and (optionally) from referenced files.

    Args:
        fh: The file-like object to read from.
        resolve (bool): Whether to resolve referenced files.

    Returns:
        Generator[PythonDependency, None, None]: A generator of PythonDependency objects.
    """
    is_temp_file = not hasattr(fh, 'name')
    path = None
    found = Path('temp_file')
    file_type = filetypes.requirements_txt
    absolute_path: Optional[Path] = None

    if not is_temp_file and is_supported_by_parser(fh.name):
        path = fh.name
        absolute_path = Path(path).resolve()
        found = absolute_path
        file_type = None

    content = fh.read()
    dependency_file = parse(content, path=path, resolve=resolve,
                            file_type=file_type)

    reqs_pkg = defaultdict(list)

    for req in dependency_file.resolved_dependencies:
        reqs_pkg[canonicalize_name(req.name)].append(req)

    for pkg, reqs in reqs_pkg.items():
        specifications = list(
            map(lambda req: parse_requirement(req, str(absolute_path)), reqs))
        version = find_version(specifications)

        yield PythonDependency(name=pkg, version=version,
                      specifications=specifications,
                      found=found,
                      absolute_path=absolute_path,
                      insecure_versions=[],
                      secure_versions=[], latest_version=None,
                      latest_version_without_known_vulnerabilities=None,
                      more_info_url=None)


def read_dependencies(fh, resolve: bool = True) -> Generator[PythonDependency, None, None]:
    """
    Reads dependencies from a file-like object.

    Args:
        fh: The file-like object to read from.
        resolve (bool): Whether to resolve referenced files.

    Returns:
        Generator[PythonDependency, None, None]: A generator of PythonDependency objects.
    """
    path = fh.name
    absolute_path = Path(path).resolve()
    found = absolute_path

    content = fh.read()
    dependency_file = parse(content, path=path, resolve=resolve)

    reqs_pkg = defaultdict(list)

    for req in dependency_file.resolved_dependencies:
        reqs_pkg[canonicalize_name(req.name)].append(req)

    for pkg, reqs in reqs_pkg.items():
        specifications = list(
            map(lambda req: parse_requirement(req, str(absolute_path)), reqs))
        version = find_version(specifications)

        yield PythonDependency(name=pkg, version=version,
                      specifications=specifications,
                      found=found,
                      absolute_path=absolute_path,
                      insecure_versions=[],
                      secure_versions=[], latest_version=None,
                      latest_version_without_known_vulnerabilities=None,
                      more_info_url=None)

def read_virtual_environment_dependencies(f: InspectableFile) -> Generator[PythonDependency, None, None]:
    """
    Reads dependencies from a virtual environment.

    Args:
        f (InspectableFile): The inspectable file representing the virtual environment.

    Returns:
        Generator[PythonDependency, None, None]: A generator of PythonDependency objects.
    """

    env_path = Path(f.file.name).resolve().parent

    if sys.platform.startswith('win'):
        site_pkgs_path = env_path / Path("Lib/site-packages/")
    else:
        site_pkgs_path = Path('lib/')
        try:
            site_pkgs_path = next((env_path / site_pkgs_path).glob("*/site-packages/"))
        except StopIteration:
            # Unable to find packages for foo env
            return

    if not site_pkgs_path.resolve().exists():
        # Unable to find packages for foo env
        return

    dep_paths = site_pkgs_path.glob("*/METADATA")

    for path in dep_paths:
        if not path.is_file():
            continue

        dist_info_folder = path.parent
        dep_name, dep_version = dist_info_folder.name.replace(".dist-info", "").split("-")

        yield PythonDependency(name=dep_name, version=dep_version,
                specifications=[
                    PythonSpecification(f"{dep_name}=={dep_version}",
                                        found=site_pkgs_path)],
                found=site_pkgs_path, insecure_versions=[],
                secure_versions=[], latest_version=None,
                latest_version_without_known_vulnerabilities=None,
                more_info_url=None)


def get_dependencies(f: InspectableFile) -> List[PythonDependency]:
    """
    Gets the dependencies for the given inspectable file.

    Args:
        f (InspectableFile): The inspectable file.

    Returns:
        List[PythonDependency]: A list of PythonDependency objects.
    """
    if not f.file_type:
        return []

    if f.file_type in [FileType.REQUIREMENTS_TXT, FileType.POETRY_LOCK,
                       FileType.PIPENV_LOCK, FileType.PYPROJECT_TOML]:
        return list(read_dependencies(f.file, resolve=True))

    if f.file_type == FileType.VIRTUAL_ENVIRONMENT:
        return list(read_virtual_environment_dependencies(f))

    return []



from functools import wraps

from safety.events.utils import emit_command_executed


def notify(func):
    """
    A decorator that wraps a function to emit events.

    Args:
      func (callable): The function to be wrapped by the decorator.

    Returns:
      callable: The wrapped function with notification logic.

    The decorator ensures that the `emit_command_executed` function is called
    after the wrapped function completes, regardless of whether it exits
    normally or via a `SystemExit` exception.

    Example:
      @notify
      def my_function(ctx, *args, **kwargs):
      # function implementation
      pass
    """

    @wraps(func)
    def inner(ctx, *args, **kwargs):
        try:
            result = func(ctx, *args, **kwargs)
            emit_command_executed(ctx.obj.event_bus, ctx, returned_code=0)
            return result
        except SystemExit as e:
            # Handle sys.exit() case
            exit_code = e.code if isinstance(e.code, int) else 1
            emit_command_executed(ctx.obj.event_bus, ctx, returned_code=exit_code)
            raise
        # Any other exceptions will bypass notification and propagate normally

    return inner



# Project options
PROJECT_INIT_CMD_NAME = "init"
PROJECT_INIT_HELP = (
    "[BETA] Used to install Safety Firewall globally, or to initialize a project in the current directory."
    "\nExample: safety init"
)
PROJECT_INIT_DIRECTORY_HELP = (
    "[BETA] Defines a directory for creating a new project. (default: current directory)\n\n"
    "[bold]Example: safety init /path/to/project[/bold]"
)



from typing import TYPE_CHECKING

from safety_schemas.models.events import EventType
from safety.events.event_bus import EventBus
from safety.events.types import InternalEventType

from .handlers import HeartbeatInspectionEventHandler

if TYPE_CHECKING:
    from safety.models import SafetyCLI


def register_event_handlers(event_bus: "EventBus", obj: "SafetyCLI") -> None:
    """
    Subscribes to the firewall events that are relevant to the current context.
    """
    handle_inspection = HeartbeatInspectionEventHandler(event_bus=event_bus)
    event_bus.subscribe([InternalEventType.EVENT_BUS_READY], handle_inspection)

    if sec_events_handler := obj.security_events_handler:
        event_bus.subscribe(
            [
                EventType.FIREWALL_CONFIGURED,
                EventType.FIREWALL_HEARTBEAT,
                EventType.FIREWALL_DISABLED,
            ],
            sec_events_handler,
        )



from abc import ABC, abstractmethod
import os
from pathlib import Path
from types import MappingProxyType
from typing import Dict, List, Optional, Optional, Tuple

from safety_schemas.models import Ecosystem, FileType


NOT_IMPLEMENTED = "You should implement this."

class FileHandler(ABC):
    """
    Abstract base class for file handlers that define how to handle specific types of files
    within an ecosystem.
    """

    def __init__(self) -> None:
        self.ecosystem: Optional[Ecosystem] = None

    def can_handle(self, root: str, file_name: str, include_files: Dict[FileType, List[Path]]) -> Optional[FileType]:
        """
        Determines if the handler can handle the given file based on its type and inclusion criteria.

        Args:
            root (str): The root directory of the file.
            file_name (str): The name of the file.
            include_files (Dict[FileType, List[Path]]): Dictionary of file types and their paths to include.

        Returns:
            Optional[FileType]: The type of the file if it can be handled, otherwise None.
        """
        # Keeping it simple for now

        if not self.ecosystem:
            return None

        for f_type in self.ecosystem.file_types:
            if f_type in include_files:
                current = Path(root, file_name).resolve()
                paths = [p.resolve() if p.is_absolute() else (root / p).resolve() for p in include_files[f_type]]
                if current in paths:
                    return f_type

            # Let's compare by name only for now
            # We can put heavier logic here, but for speed reasons,
            # right now is very basic, we will improve this later.
            # Custom matching per File Type
            if file_name.lower().endswith(f_type.value.lower()):
                return f_type

        return None

    @abstractmethod
    def download_required_assets(self, session) -> Dict[str, str]:
        """
        Abstract method to download required assets for handling files. Should be implemented
        by subclasses.

        Args:
            session: The session object for making network requests.

        Returns:
            Dict[str, str]: A dictionary of downloaded assets.
        """
        return NotImplementedError(NOT_IMPLEMENTED)


class PythonFileHandler(FileHandler):
    """
    Handler for Python files within the Python ecosystem.
    """
    # Example of a Python File Handler

    def __init__(self) -> None:
        super().__init__()
        self.ecosystem = Ecosystem.PYTHON

    def download_required_assets(self, session) -> None:
        """
        Downloads the required assets for handling Python files, specifically the Safety database.

        Args:
            session: The session object for making network requests.
        """
        from safety.safety import fetch_database

        SAFETY_DB_DIR = os.getenv("SAFETY_DB_DIR")

        db = False if SAFETY_DB_DIR is None else SAFETY_DB_DIR

        # Fetch both the full and partial Safety databases
        fetch_database(session=session, full=False, db=db, cached=True,
                       telemetry=True, ecosystem=Ecosystem.PYTHON,
                       from_cache=False)

        fetch_database(session=session, full=True, db=db, cached=True,
                                telemetry=True, ecosystem=Ecosystem.PYTHON,
                                from_cache=False)


class SafetyProjectFileHandler(FileHandler):
    """
    Handler for Safety project files within the Safety project ecosystem.
    """
    # Example of a Python File Handler

    def __init__(self) -> None:
        super().__init__()
        self.ecosystem = Ecosystem.SAFETY_PROJECT

    def download_required_assets(self, session) -> None:
        """
        No required assets to download for Safety project files.
        """
        pass

# Mapping of ecosystems to their corresponding file handlers
ECOSYSTEM_HANDLER_MAPPING = MappingProxyType({
    Ecosystem.PYTHON: PythonFileHandler,
    Ecosystem.SAFETY_PROJECT: SafetyProjectFileHandler,
})



import logging
import sys

import typer
from rich.prompt import Prompt

from safety.console import main_console as console
from safety.decorators import notify
from safety.events.utils import emit_firewall_disabled


# TODO: refactor this import and the related code
# For now, let's keep it as is
from safety.scan.constants import DEFAULT_EPILOG
from safety.error_handlers import handle_cmd_exception

from ..cli_util import (
    CommandType,
    FeatureType,
    SafetyCLICommand,
    SafetyCLISubGroup,
    pass_safety_cli_obj,
)
from ..constants import CONTEXT_COMMAND_TYPE, CONTEXT_FEATURE_TYPE, EXIT_CODE_OK
from ..tool.interceptors import create_interceptor
from ..tool.main import reset_system
from .constants import (
    FIREWALL_CMD_NAME,
    FIREWALL_HELP,
    MSG_FEEDBACK,
    MSG_REQ_FILE_LINE,
    MSG_UNINSTALL_EXPLANATION,
    MSG_UNINSTALL_PIP_ALIAS,
    MSG_UNINSTALL_PIP_CONFIG,
    MSG_UNINSTALL_SUCCESS,
    UNINSTALL_CMD_NAME,
    UNINSTALL_HELP,
)


firewall_app = typer.Typer(
    rich_markup_mode="rich", cls=SafetyCLISubGroup, name=FIREWALL_CMD_NAME
)


LOG = logging.getLogger(__name__)


init_app = typer.Typer(rich_markup_mode="rich", cls=SafetyCLISubGroup)


@firewall_app.callback(
    cls=SafetyCLISubGroup,
    help=FIREWALL_HELP,
    epilog=DEFAULT_EPILOG,
    context_settings={
        "allow_extra_args": True,
        "ignore_unknown_options": True,
        CONTEXT_COMMAND_TYPE: CommandType.BETA,
        CONTEXT_FEATURE_TYPE: FeatureType.FIREWALL,
    },
)
@pass_safety_cli_obj
def firewall(ctx: typer.Context) -> None:
    """
    Main callback for the firewall commands.

    Args:
        ctx (typer.Context): The Typer context object.
    """
    LOG.info("firewall callback started")


@firewall_app.command(
    cls=SafetyCLICommand,
    name=UNINSTALL_CMD_NAME,
    help=UNINSTALL_HELP,
    options_metavar="[OPTIONS]",
    context_settings={
        "allow_extra_args": True,
        "ignore_unknown_options": True,
        CONTEXT_COMMAND_TYPE: CommandType.BETA,
        CONTEXT_FEATURE_TYPE: FeatureType.FIREWALL,
    },
)
@handle_cmd_exception
@notify
def uninstall(ctx: typer.Context):
    console.print()
    console.print(MSG_UNINSTALL_EXPLANATION)

    console.print()
    prompt = "Uninstall?"
    should_uninstall = (
        Prompt.ask(
            prompt=prompt,
            choices=["y", "n"],
            default="y",
            show_default=True,
            console=console,
        ).lower()
        == "y"
    )

    if not should_uninstall:
        sys.exit(EXIT_CODE_OK)

    console.print()
    console.print(MSG_UNINSTALL_PIP_CONFIG)
    # TODO: Make it robust. The reset per tool should be included in remove
    # interceptors
    reset_system()

    # TODO: support reset project files

    console.print(MSG_UNINSTALL_PIP_ALIAS)
    interceptor = create_interceptor()
    interceptor.remove_interceptors()

    console.print()
    console.print(MSG_UNINSTALL_SUCCESS)

    console.print()
    console.print(MSG_REQ_FILE_LINE)

    console.print()

    console.print(MSG_FEEDBACK)

    console.print()
    prompt = "Feedback (or enter to exit)"
    feedback = Prompt.ask(prompt)
    feedback = None if len(feedback) <= 0 else feedback

    emit_firewall_disabled(event_bus=ctx.obj.event_bus, reason=feedback)

    if feedback:
        console.print()
        console.print("Thank you for your feedback!")
